"""Generate student & autograder views of a notebook in okpy format."""

import copy
import json
import nbformat
import pprint
import os
import re
import shutil
import subprocess
import yaml

from collections import namedtuple

NB_VERSION = 4
BLOCK_QUOTE = "```"
COMMENT_PREFIX = "#"
TEST_HEADERS = ["TEST", "HIDDEN TEST"]
ALLOWED_NAME = re.compile(r'[A-Za-z][A-Za-z0-9_]*')
NB_VERSION = 4


def convert_to_ok(nb_path, dir, args):
    """Convert a master notebook to an ok notebook, tests dir, and .ok file.

    nb -- Path
    endpoint -- OK endpoint for notebook submission (e.g., cal/data100/sp19)
    dir -- Path
    """
    ok_nb_path = dir / nb_path.name
    tests_dir = dir / 'tests'
    os.makedirs(tests_dir, exist_ok=True)

    # copy files
    for file in args.files:
        shutil.copy(file, str(dir))

    with open(nb_path) as f:
        nb = nbformat.read(f, NB_VERSION)
    ok_cells = gen_ok_cells(nb['cells'], tests_dir)

    if not args.no_init_cell:
        init = gen_init_cell()
        nb['cells'] = [init] + ok_cells
    else:
        nb['cells'] = ok_cells

    if not args.no_check_all:
        nb['cells'] += gen_check_all_cell()
    if not args.no_export_cell:
        nb['cells'] += gen_export_cells(nb_path, args.instructions, filtering = not args.no_filter)
        
    remove_output(nb)

    with open(ok_nb_path, 'w') as f:
        nbformat.write(nb, f, NB_VERSION)
    return ok_nb_path


def gen_init_cell():
    """Generate a cell to initialize ok object."""
    cell = nbformat.v4.new_code_cell("# Initialize Otter\nimport otter\ngrader = otter.Notebook()")
    lock(cell)
    return cell


def gen_export_cells(nb_path, instruction_text, filtering=True):
    """Generate submit cells."""
    instructions = nbformat.v4.new_markdown_cell()
    instructions.source = "## Submission\n\nMake sure you have run all cells in your notebook in order before \
    running the cell below, so that all images/graphs appear in the output. **Please save before submitting!**"
    
    if instruction_text:
        instructions.source += '\n\n' + instruction_text

    export = nbformat.v4.new_code_cell()
    source_lines = ["# Save your notebook first, then run this cell to export."]
    if filtering:
        source_lines.append(f"grader.export(\"{ nb_path.name }\")")
    else:
        source_lines.append(f"grader.export(\"{ nb_path.name }\", filtering=False)")
    export.source = "\n".join(source_lines)

    lock(instructions)
    lock(export)

    return [instructions, export, nbformat.v4.new_markdown_cell(" ")]     # last cell is buffer


def gen_check_all_cell():
    """Generate submit cells."""
    instructions = nbformat.v4.new_markdown_cell()
    instructions.source = "To double-check your work, the cell below will rerun all of the autograder tests."

    check_all = nbformat.v4.new_code_cell("grader.check_all()")

    lock(instructions)
    lock(check_all)

    return [instructions, check_all]


def gen_ok_cells(cells, tests_dir):
    """Generate notebook cells for the OK version of a master notebook.

    Returns:
        (ok_cells, list of manual question names)
    """
    ok_cells = []
    question = {}
    processed_response = False
    tests = []
    # hidden_tests = []
    manual_questions = []
    need_close_export = False

    for cell in cells:

        # this is the prompt cell or if a manual question then the solution cell
        if question and not processed_response:
            assert not is_question_cell(cell), cell
            assert not is_test_cell(cell), cell
            # assert not is_solution_cell(cell) or is_markdown_solution_cell(cell), cell

            # # if this isn't a MD solution cell but in a manual question, it has a prompt
            # if not is_solution_cell(cell) and question.get('manual', False):
            #     md_has_prompt = True
            
            # # if there is no prompt, add a prompt cell
            # elif is_markdown_solution_cell(cell) and not md_has_prompt:
            #     ok_cells.append(nbformat.v4.new_markdown_cell(MD_ANSWER_CELL_TEMPLATE))

            ok_cells.append(cell)
            processed_response = True

        # if this is a test cell, parse and add to correct group
        elif question and processed_response and is_test_cell(cell):
            test = read_test(cell)
            # if test.hidden:
            #     hidden_tests.append(test)
            # else:
            tests.append(test)

        # # if this is a solution cell, append. if manual question and no prompt, also append prompt cell
        # elif question and processed_response and is_solution_cell(cell):
        #     if is_markdown_solution_cell(cell) and not md_has_prompt:
        #         ok_cells.append(nbformat.v4.new_markdown_cell(MD_ANSWER_CELL_TEMPLATE))
        #     ok_cells.append(cell)

        else:
            # the question is over
            if question and processed_response:
                if tests:
                    ok_cells.append(gen_test_cell(question, tests, tests_dir))
                # if hidden_tests:
                #     gen_test_cell(question, hidden_tests, tests_dir, hidden=True)

                # add a cell with <!-- END QUESTION --> if a manually graded question
                manual = question.get('manual', False)
                if manual:
                    # ok_cells.append(gen_close_export_cell())
                    need_close_export = True
                
                question, processed_response, tests, hidden_tests = {}, False, [], []

            if is_question_cell(cell):
                question = read_question_metadata(cell)
                manual = question.get('manual', False)
                format = question.get('format', '')
                in_manual_block = False
                if manual and need_close_export:
                    manual_questions.append(question['name'])
                    in_manual_block = True
                elif need_close_export:
                    add_close_export_to_cell(cell)
                    need_close_export = False
                ok_cells.append(gen_question_cell(cell, manual, format, in_manual_block))

            # elif is_solution_cell(cell):
            #     if is_markdown_solution_cell(cell):
            #         ok_cells.append(nbformat.v4.new_markdown_cell(MD_ANSWER_CELL_TEMPLATE))
            #     ok_cells.append(cell)

            else:
                assert not is_test_cell(cell), 'Test outside of a question: ' + str(cell)

                if need_close_export:
                    if cell['cell_type'] == 'code':
                        ok_cells.append(gen_close_export_cell())
                    else:
                        add_close_export_to_cell(cell)
                    need_close_export = False

                ok_cells.append(cell)

    if tests:
        ok_cells.append(gen_test_cell(question, tests, tests_dir))
    # if hidden_tests:
    #     gen_test_cell(question, hidden_tests, tests_dir, hidden=True)

    return ok_cells


def get_source(cell):
    """Get the source code of a cell in a way that works for both nbformat and json."""
    source = cell['source']
    if isinstance(source, str):
        return cell['source'].split('\n')
    elif isinstance(source, list):
        return [line.strip('\n') for line in source]
    assert 'unknown source type', type(source)


def is_question_cell(cell):
    """Whether cell contains BEGIN QUESTION in a block quote."""
    if cell['cell_type'] != 'markdown':
        return False
    return find_question_spec(get_source(cell)) is not None


def find_question_spec(source):
    """Return line number of the BEGIN QUESTION line or None."""
    block_quotes = [i for i, line in enumerate(source) if
                    line == BLOCK_QUOTE]
    assert len(block_quotes) % 2 == 0, 'cannot parse ' + str(source)
    begins = [block_quotes[i] + 1 for i in range(0, len(block_quotes), 2) if
              source[block_quotes[i]+1].strip(' ') == 'BEGIN QUESTION']
    assert len(begins) <= 1, 'multiple questions defined in ' + str(source)
    return begins[0] if begins else None


def gen_question_cell(cell, manual, format, in_manual_block):
    """Return the cell with metadata hidden in an HTML comment."""
    cell = copy.deepcopy(cell)
    source = get_source(cell)
    if manual and not in_manual_block:
        source = ["<!-- BEGIN QUESTION -->", ""] + source
    begin_question_line = find_question_spec(source)
    start = begin_question_line - 1
    assert source[start].strip() == BLOCK_QUOTE
    end = begin_question_line
    while source[end].strip() != BLOCK_QUOTE:
        end += 1
    source[start] = "<!--"
    source[end] = "-->"
    cell['source'] = '\n'.join(source)
    lock(cell)
    return cell


def gen_close_export_cell():
    """Returns a new cell to end question export"""
    cell = nbformat.v4.new_markdown_cell("<!-- END QUESTION -->")
    lock(cell)
    return cell


def add_close_export_to_cell(cell):
    """Adds an export close to the top of the cell"""
    source = get_source(cell)
    source = ["<!-- END QUESTION -->\n", "\n"] + source
    cell['source'] = source


def read_question_metadata(cell):
    """Return question metadata from a question cell."""
    source = get_source(cell)
    begin_question_line = find_question_spec(source)
    i, lines = begin_question_line + 1, []
    while source[i].strip() != BLOCK_QUOTE:
        lines.append(source[i])
        i = i + 1
    metadata = yaml.full_load('\n'.join(lines))
    assert ALLOWED_NAME.match(metadata.get('name', '')), metadata
    return metadata


def is_test_cell(cell):
    """Return whether it's a code cell containing a test."""
    if cell['cell_type'] != 'code':
        return False
    source = get_source(cell)
    delimiters = COMMENT_PREFIX + ' \n'
    return source and source[0].strip(delimiters) in TEST_HEADERS


Test = namedtuple('Test', ['input', 'output', 'hidden'])


def read_test(cell):
    """Return the contents of a test as an (input, output, hidden) tuple."""
    hidden = 'HIDDEN' in get_source(cell)[0]
    output = ''
    for o in cell['outputs']:
        output += ''.join(o.get('text', ''))
        results = o.get('data', {}).get('text/plain')
        if results and isinstance(results, list):
            output += results[0]
        elif results:
            output += results
    return Test('\n'.join(get_source(cell)[1:]), output, hidden)


def write_test(path, test):
    """Write an OK test file."""
    with open(path, 'w') as f:
        f.write('test = ')
        pprint.pprint(test, f, indent=4, width=200, depth=None)


def gen_test_cell(question, tests, tests_dir):
    """Return a test cell."""
    cell = nbformat.v4.new_code_cell()
    # if hidden:
    #     cell.source = ['grader.check("{}")'.format(question['name'] + "H")]
    # else:
    cell.source = ['grader.check("{}")'.format(question['name'])]
    suites = [gen_suite(tests)]

    # # if both keys, use those values
    # if 'public_points' in question and 'private_points' in question:
    #     if hidden:
    #         points = question['private_points']
    #     else:
    #         points = question['public_points']

    # # if only public points, assume public & private worth same
    # elif 'public_points' in question:
    #     points = question['public_points']

    # # if only private, assume public worth 0 points
    # elif 'private_points' in question:
    #     if hidden:
    #         points = question['private_points']
    #     else:
    #         points = 0

    # # else get points and default to 1 if not presnet
    # else:
    points = question.get('points', 1)
    
    test = {
        'name': question['name'],
        'points': points,
        # 'hidden': hidden,
        'suites': suites,
    }
    # if hidden:
    #     write_test(tests_dir / (question['name'] + "H" + '.py'), test)
    # else:
    write_test(tests_dir / (question['name'] + '.py'), test)
    lock(cell)
    return cell


def gen_suite(tests):
    """Generate an ok test suite for a test."""
    cases = [gen_case(test) for test in tests]
    return  {
      'cases': cases,
      'scored': True,
      'setup': '',
      'teardown': '',
      'type': 'doctest'
    }


def gen_case(test):
    """Generate an ok test case for a test."""
    # TODO(denero) This should involve a Python parser, but it doesn't...
    code_lines = []
    last_end_escape = False
    for line in test.input.split('\n'):
        if re.match(r"\s", line) or last_end_escape:
            code_lines.append('... ' + line)
        else:
            code_lines.append('>>> ' + line)
        last_end_escape = line.endswith("\\")
    # Suppress intermediate output from evaluation
    for i in range(len(code_lines) - 1):
        if code_lines[i+1].startswith('>>>') and len(code_lines[i].strip()) > 3 and not code_lines[i].endswith("\\"):
            code_lines[i] += ';'
    code_lines.append(test.output)
    return {
        'code': '\n'.join(code_lines),
        'hidden': test.hidden,
        'locked': False
    }


solution_assignment_re = re.compile('(\\s*[a-zA-Z0-9_ ]*=)(.*) #[ ]?SOLUTION')
def solution_assignment_sub(match):
    prefix = match.group(1)
    # sol = match.group(2)
    return prefix + ' ...'


solution_line_re = re.compile('(\\s*)([^#\n]+) #[ ]?SOLUTION')
def solution_line_sub(match):
    prefix = match.group(1)
    return prefix + '...'


text_solution_line_re = re.compile(r'\s*\*\*SOLUTION:?\*\*:?.*')
begin_solution_re = re.compile(r'(\s*)# BEGIN SOLUTION( NO PROMPT)?')
skip_suffixes = ['# SOLUTION NO PROMPT', '# BEGIN PROMPT', '# END PROMPT']


SUBSTITUTIONS = [
    (solution_assignment_re, solution_assignment_sub),
    (solution_line_re, solution_line_sub),
]


def replace_solutions(lines):
    """Replace solutions in lines, a list of strings."""
    if text_solution_line_re.match(lines[0]):
        return ['*Write your answer here, replacing this text.*']
    stripped = []
    solution = False
    for line in lines:
        if any(line.endswith(s) for s in skip_suffixes):
            continue
        if solution and not line.endswith('# END SOLUTION'):
            continue
        if line.endswith('# END SOLUTION'):
            assert solution, 'END SOLUTION without BEGIN SOLUTION in ' + str(lines)
            solution = False
            continue
        begin_solution = begin_solution_re.match(line)
        if begin_solution:
            assert not solution, 'Nested BEGIN SOLUTION in ' + str(lines)
            solution = True
            if not begin_solution.group(2):
                line = begin_solution.group(1) + '...'
            else:
                continue
        for exp, sub in SUBSTITUTIONS:
            m = exp.match(line)
            if m:
                line = sub(m)
        stripped.append(line)
    assert not solution, 'BEGIN SOLUTION without END SOLUTION in ' + str(lines)
    return stripped


def strip_solutions(original_nb_path, stripped_nb_path):
    """Write a notebook with solutions stripped."""
    with open(original_nb_path) as f:
        nb = nbformat.read(f, NB_VERSION)
    for cell in nb['cells']:
        cell['source'] = '\n'.join(replace_solutions(get_source(cell)))
    with open(stripped_nb_path, 'w') as f:
        nbformat.write(nb, f, NB_VERSION)


def remove_output(nb):
    """Remove all outputs."""
    for cell in nb['cells']:
        if 'outputs' in cell:
            cell['outputs'] = []


def lock(cell):
    m = cell['metadata']
    m["editable"] = False
    m["deletable"] = False


def remove_hidden_tests(test_dir):
    """Rewrite test files to remove hidden tests."""
    for f in test_dir.iterdir():
        if f.name == '__init__.py' or f.suffix != '.py':
            continue
        locals = {}
        with open(f) as f2:
            exec(f2.read(), globals(), locals)
        test = locals['test']
        for suite in test['suites']:
            for i, case in list(enumerate(suite['cases']))[::-1]:
                if case['hidden']:
                    suite['cases'].pop(i)
        write_test(f, test)


def gen_views(master_nb, result_dir, args):
    """Generate student and autograder views.

    master_nb -- Dict of master notebook JSON
    result_dir -- Path to the result directory
    """
    autograder_dir = result_dir / 'autograder'
    student_dir = result_dir / 'student'
    os.makedirs(autograder_dir, exist_ok=True)
    ok_nb_path = convert_to_ok(master_nb, autograder_dir, args)
    shutil.rmtree(student_dir, ignore_errors=True)
    shutil.copytree(autograder_dir, student_dir)
    student_nb_path = student_dir / ok_nb_path.name
    os.remove(student_nb_path)
    strip_solutions(ok_nb_path, student_nb_path)
    remove_hidden_tests(student_dir / 'tests')



#################################
##### Tests for otter grade #####
#################################

import os
import unittest
import subprocess
import json
import re
import pandas as pd

from unittest import mock
from subprocess import PIPE
from glob import glob

from otter.grade import main as grade
from otter.metadata import GradescopeParser, CanvasParser, JSONParser, YAMLParser

# read in argument parser
bin_globals = {}

with open("bin/otter") as f:
    exec(f.read(), bin_globals)

parser = bin_globals["parser"]

TEST_FILES_PATH = "test/test-grade/"

class TestGrade(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        print("\n\n\n" + ("=" * 60) + f"\nRunning {__name__}.{cls.__name__}\n" + ("=" * 60) + "\n")

        create_image_cmd = ["make", "docker-test"]
        create_image = subprocess.run(create_image_cmd, stdout=PIPE, stderr=PIPE)
        assert not create_image.stderr, create_image.stderr.decode("utf-8")
    
    def setUp(self):
        """
        Load in point values
        """
        self.test_points = {}
        for test_file in glob(TEST_FILES_PATH + "tests/*.py"):
            env = {}
            with open(test_file) as f:
                exec(f.read(), env)
            self.test_points[env['test']['name']] = env['test']['points']


    def test_docker(self):
        """
        Check that we have the right container installed and that docker is running
        """
        # use docker image inspect to see that the image is installed and tagged as otter-grader
        inspect = subprocess.run(["docker", "image", "inspect", "otter-test"], stdout=PIPE, stderr=PIPE)

        # assert that it didn't fail, it will fail if it is not installed
        self.assertEqual(len(inspect.stderr), 0, inspect.stderr.decode("utf-8"))


    def test_metadata_parsers(self):
        """
        Check that metadata parsers work correctly
        """
        correct_metadata = [
            {
                "identifier": "12345",
                "filename": "12345_empty_file.ipynb"
            }, {
                "identifier": "23456",
                "filename": "23456_empty_file.ipynb"
            }, {
                "identifier": "34567",
                "filename": "34567_empty_file.ipynb"
            }, {
                "identifier": "45678",
                "filename": "45678_empty_file.ipynb"
            }, {
                "identifier": "56789",
                "filename": "56789_empty_file.ipynb"
            }
        ]

        correct_file_to_id = {
            "12345_empty_file.ipynb": "12345",
            "23456_empty_file.ipynb": "23456",
            "34567_empty_file.ipynb": "34567",
            "45678_empty_file.ipynb": "45678",
            "56789_empty_file.ipynb": "56789",
        }

        correct_id_to_file = {
            "12345": "12345_empty_file.ipynb",
            "23456": "23456_empty_file.ipynb",
            "34567": "34567_empty_file.ipynb",
            "45678": "45678_empty_file.ipynb",
            "56789": "56789_empty_file.ipynb",
        }

        correct_filenames = [
            "12345_empty_file.ipynb",
            "23456_empty_file.ipynb",
            "34567_empty_file.ipynb",
            "45678_empty_file.ipynb",
            "56789_empty_file.ipynb",
        ]

        correct_identifiers = [
            "12345",
            "23456",
            "34567",
            "45678",
            "56789",
        ]

        try:
            # gradescope parser
            gs_parser = GradescopeParser(TEST_FILES_PATH + "gradescope-export")
            self.assertCountEqual(gs_parser.get_metadata(), correct_metadata)
            self.assertCountEqual(gs_parser.get_filenames(), correct_filenames)
            self.assertCountEqual(gs_parser.get_identifiers(), correct_identifiers)
            for file, identifier in zip(correct_file_to_id, correct_id_to_file):
                self.assertEqual(correct_file_to_id[file], gs_parser.file_to_id(file))
                self.assertEqual(correct_id_to_file[identifier], gs_parser.id_to_file(identifier))

            # canvas parser
            canvas_parser = CanvasParser(TEST_FILES_PATH + "canvas-export")
            self.assertCountEqual(canvas_parser.get_metadata(), correct_metadata)
            self.assertCountEqual(canvas_parser.get_filenames(), correct_filenames)
            self.assertCountEqual(canvas_parser.get_identifiers(), correct_identifiers)
            for file, identifier in zip(correct_file_to_id, correct_id_to_file):
                self.assertEqual(correct_file_to_id[file], canvas_parser.file_to_id(file))
                self.assertEqual(correct_id_to_file[identifier], canvas_parser.id_to_file(identifier))

            # JSON parser
            json_parser = JSONParser(TEST_FILES_PATH + "meta.json")
            self.assertCountEqual(json_parser.get_metadata(), correct_metadata)
            self.assertCountEqual(json_parser.get_filenames(), correct_filenames)
            self.assertCountEqual(json_parser.get_identifiers(), correct_identifiers)
            for file, identifier in zip(correct_file_to_id, correct_id_to_file):
                self.assertEqual(correct_file_to_id[file], json_parser.file_to_id(file))
                self.assertEqual(correct_id_to_file[identifier], json_parser.id_to_file(identifier))

            # YAML parser
            yaml_parser = YAMLParser(TEST_FILES_PATH + "meta.yml")
            self.assertCountEqual(yaml_parser.get_metadata(), correct_metadata)
            self.assertCountEqual(yaml_parser.get_filenames(), correct_filenames)
            self.assertCountEqual(yaml_parser.get_identifiers(), correct_identifiers)
            for file, identifier in zip(correct_file_to_id, correct_id_to_file):
                self.assertEqual(correct_file_to_id[file], yaml_parser.file_to_id(file))
                self.assertEqual(correct_id_to_file[identifier], yaml_parser.id_to_file(identifier))

            # cleanup
            gs_rm = subprocess.run(["rm", "-rf"] + glob(TEST_FILES_PATH + "gradescope-export/*.ipynb"), stdout=PIPE, stderr=PIPE)
            self.assertEqual(len(gs_rm.stderr), 0, gs_rm.stderr.decode("utf-8"))

        except:
            # cleanup
            gs_rm = subprocess.run(["rm", "-rf"] + glob(TEST_FILES_PATH + "gradescope-export/*.ipynb"), stdout=PIPE, stderr=PIPE)
            self.assertEqual(len(gs_rm.stderr), 0, gs_rm.stderr.decode("utf-8"))
            raise


    def test_notebooks(self):
        """
        Check that the example of 100 notebooks runs correctely locally.
        """
        # grade the 100 notebooks
        grade_command = ["grade",
            # NO METADATA PASSED, test case when no metadata provided
            # "-y", TEST_FILES_PATH + "notebooks/meta.yml", 
            "-p", TEST_FILES_PATH + "notebooks/", 
            "-t", TEST_FILES_PATH + "tests/", 
            "-r", TEST_FILES_PATH + "requirements.txt",
            "-o", "test/",
            "--image", "otter-test",
            "-v"
        ]
        args = parser.parse_args(grade_command)
        args.func = grade
        args.func(args)

        # read the output and expected output
        df_test = pd.read_csv("test/final_grades.csv")
        self.assertTrue("identifier" not in df_test.columns, "did not drop identifier column when no metadata passed")

        # sort by filename
        df_test = df_test.sort_values("file").reset_index(drop=True)
        df_test["failures"] = df_test["file"].apply(lambda x: [int(n) for n in re.split(r"\D+", x) if len(n) > 0])

        # check point values
        for _, row in df_test.iterrows():
            for test in self.test_points:
                if int(re.sub(r"\D", "", test)) in row["failures"]:
                    self.assertEqual(row[test], 0, "{} supposed to fail {} but passed".format(row["file"], test))
                else:
                    self.assertEqual(row[test], self.test_points[test], "{} supposed to pass {} but failed".format(row["file"], test))

        # df_correct = pd.read_csv(TEST_FILES_PATH + "final_grades_correct_notebooks.csv").sort_values("identifier").reset_index(drop=True)

        # # assert the dataframes are as expected
        # self.assertTrue(df_test.equals(df_correct), "Dataframes not equal")

        # remove the extra output
        cleanup_command = ["rm", "test/final_grades.csv"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)

        # assert cleanup worked
        self.assertEqual(len(cleanup.stderr), 0, "Error in cleanup")


    def test_notebooks_with_pdfs(self):
        """
        Check that the example of 100 notebooks runs correctely locally.
        """
        # grade the 100 notebooks
        grade_command = ["grade",
            "-y", TEST_FILES_PATH + "notebooks/meta.yml", 
            "-p", TEST_FILES_PATH + "notebooks/", 
            "-t", TEST_FILES_PATH + "tests/", 
            "-r", TEST_FILES_PATH + "requirements.txt",
            "-o", "test/",
            "--pdfs", "tags",
            "--containers", "5",
            "--image", "otter-test"
        ]
        args = parser.parse_args(grade_command)
        args.func = grade
        args.func(args)

        # check that we have PDFs
        self.assertTrue(os.path.isdir("test/submission_pdfs"))
        for file in glob(TEST_FILES_PATH + "notebooks/*.ipynb"):
            pdf = "test/submission_pdfs/" + os.path.split(file)[1][:-5] + "pdf"
            self.assertTrue(os.path.isfile(pdf))

        # remove the extra output
        cleanup_command = ["rm", "-rf", "test/final_grades.csv", "test/submission_pdfs"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(cleanup.stderr), 0, cleanup.stderr.decode("utf-8"))


    def test_scripts(self):
        """
        Check that the example of 100 scripts runs correctely locally.
        """
        grade_command = ["grade",
            "-sy", TEST_FILES_PATH + "scripts/meta.yml", 
            "-p", TEST_FILES_PATH + "scripts/", 
            "-t", TEST_FILES_PATH + "tests/", 
            "-r", TEST_FILES_PATH + "requirements.txt",
            "-o", "test/",
            "--image", "otter-test"
        ]
        args = parser.parse_args(grade_command)
        args.func = grade
        args.func(args)

        # read the output and expected output
        df_test = pd.read_csv("test/final_grades.csv").sort_values("identifier").reset_index(drop=True)
        df_test["failures"] = df_test["identifier"].apply(lambda x: [int(n) for n in re.split(r"\D+", x) if len(n) > 0])

        # check point values
        for _, row in df_test.iterrows():
            for test in self.test_points:
                if int(re.sub(r"\D", "", test)) in row["failures"]:
                    self.assertEqual(row[test], 0, "{} supposed to fail {} but passed".format(row["identifier"], test))
                else:
                    self.assertEqual(row[test], self.test_points[test], "{} supposed to pass {} but failed".format(row["identifier"], test))
        
        # grade the 100 scripts
        
        # args = parser.parse_args(grade_command)
        # args.func = grade
        # args.func(args)

        # # read the output and expected output
        # df_test = pd.read_csv("test/final_grades.csv").sort_values("identifier").reset_index(drop=True)
        # df_correct = pd.read_csv(TEST_FILES_PATH + "final_grades_correct_script.csv").sort_values("identifier").reset_index(drop=True)

        # # assert the dataframes are as expected
        # self.assertTrue(df_test.equals(df_correct), "Dataframes not equal")

        # remove the extra output
        cleanup_command = ["rm", "test/final_grades.csv"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)

        # assert cleanup worked
        self.assertEqual(len(cleanup.stderr), 0, "Error in cleanup")



########################################################
##### Docker Container Management for Otter-Grader #####
########################################################

import subprocess
import re
import os
import shutil
import tempfile
import pandas as pd

from subprocess import PIPE
from concurrent.futures import ThreadPoolExecutor, wait

from .metadata import GradescopeParser

def launch_parallel_containers(tests_dir, notebooks_dir, verbose=False, pdfs=None, reqs=None, 
    num_containers=None, image="ucbdsinfra/otter-grader", scripts=False, no_kill=False, output_path="./", 
    debug=False, seed=None, zips=False, meta_parser=None):
    """Grades notebooks in parallel docker containers

    This function runs ``num_containers`` docker containers in parallel to grade the student submissions
    in ``notebooks_dir`` using the tests in ``tests_dir``. It can additionally generate PDFs for the parts
    of the assignment needing manual grading. 

    Args:
        tests_dir (``str``): path to directory of tests
        notebooks_dir (``str``): path to directory of student submissions to be graded
        verbose (``bool``, optional): whether status messages should be printed to the command line
        pdfs (``str``, optional): string indicating what type of PDF to generate; defaults to ``None``
            indicating no PDFS but can be one of ``["unfiltered", "tags", "html"]`` corresponding to
            the type of cell filtering to perform upon generation
        reqs (``str``, optional): Path to requirements.txt file with non-standard packages needed
        num_containers (``int``, optional): The number of parallel containers that will be run
        image (``str``, optional): a docker image tag to be used for grading environment
        scripts (``bool``, optional): whether student submissions are Python scripts rather than
            Jupyter notebooks
        no_kill (``bool``, optional): whether the grading containers should be kept running after
            grading finishes
        output_path (``str``, optional): path at which to write grades CSVs copied from the container
        debug (``bool``, False): whether to run grading in debug mode (prints grading STDOUT and STDERR 
            from each container to the command line)
        seed (``int``, optional): a random seed to use for intercell seeding during grading
        zips (``bool``, False): whether the submissions are zip files formatted from ``Notebook.export``
        meta_parser (object, optional): a metadata parser instance; one of the classes defined in
            ``otter.metadata``

    Returns:
        ``list`` of ``pandas.core.frame.DataFrame``: the grades returned by each container spawned during
            grading
    """
    if not num_containers:
        num_containers = 4

    # list all notebooks in the dir
    dir_path = os.path.abspath(notebooks_dir)
    file_extension = (".zip", ".py", ".ipynb")[[zips, scripts, True].index(True)]
    notebooks = [os.path.join(dir_path, f) for f in os.listdir(dir_path) \
        if os.path.isfile(os.path.join(dir_path, f)) and f.endswith(file_extension)]

    # fix number of containers (overriding user input if necessary)
    if len(notebooks) < num_containers:
        num_containers = len(notebooks)

    # # calculate number of notebooks per container
    # num_per_group = int(len(notebooks) / num_containers)

    try:
        # create temp directories and add non-notebook files
        dirs = [tempfile.mkdtemp() for _ in range(num_containers)]

        # if GS export, get list of folder names to ignore when copying files
        if isinstance(meta_parser, GradescopeParser):
            ignore_folders = meta_parser.get_folders()
        else:
            ignore_folders = []

        # copy all non-notebook files into each temp directory
        for i in range(num_containers):
            for file in os.listdir(dir_path):
                fp = os.path.join(dir_path, file)
                if os.path.isfile(fp) and os.path.splitext(fp)[1] != file_extension:
                    shutil.copy(fp, dirs[i])
                elif os.path.isdir(fp) and fp not in ignore_folders:                    
                    shutil.copytree(fp, os.path.join(dirs[i], file))

        # copy notebooks in tmp directories
        for k, v in enumerate(notebooks):
            shutil.copy(v, dirs[k % num_containers])

        # execute containers in parallel
        pool = ThreadPoolExecutor(num_containers)
        futures = []
        for i in range(num_containers):
            futures += [pool.submit(grade_assignments, 
                tests_dir, 
                dirs[i], 
                str(i), 
                verbose=verbose, 
                pdfs=pdfs,
                # unfiltered_pdfs=unfiltered_pdfs, 
                # tag_filter=tag_filter,
                # html_filter=html_filter,
                reqs=reqs,
                image=image,
                scripts=scripts,
                no_kill=no_kill,
                output_path=output_path,
                debug=debug,
                seed=seed,
                zips=zips
            )]

        # stop execution while containers are running
        finished_futures = wait(futures)
    
    # cleanup temp directories on failure
    except:
        for i in range(num_containers):
            shutil.rmtree(dirs[i], ignore_errors=True)
        raise
    
    # cleanup temp directories
    else:
        for i in range(num_containers):
            shutil.rmtree(dirs[i])

    # return list of dataframes
    return [df.result() for df in finished_futures[0]]

def grade_assignments(tests_dir, notebooks_dir, id, image="ucbdsinfra/otter-grader", verbose=False, 
    pdfs=False, reqs=None, scripts=False, no_kill=False, output_path="./", debug=False, seed=None, zips=False):
    """
    Grades multiple assignments in a directory using a single docker container. If no PDF assignment is 
    wanted, set all three PDF params (``unfiltered_pdfs``, ``tag_filter``, and ``html_filter``) to ``False``.

    Args:
        tests_dir (``str``): path to directory of tests
        notebooks_dir (``str``): path to directory of student submissions to be graded
        id (``int``): grading container index
        image (``str``, optional): a docker image tag to be used for grading environment
        verbose (``bool``, optional): whether status messages should be printed to the command line
        pdfs (``str``, optional): string indicating what type of PDF to generate; defaults to ``None``
            indicating no PDFS but can be one of ``["unfiltered", "tags", "html"]`` corresponding to
            the type of cell filtering to perform upon generation
        reqs (``str``, optional): Path to requirements.txt file with non-standard packages needed
        scripts (``bool``, optional): whether student submissions are Python scripts rather than
            Jupyter notebooks
        no_kill (``bool``, optional): whether the grading containers should be kept running after
            grading finishes
        output_path (``str``, optional): path at which to write grades CSVs copied from the container
        debug (``bool``, False): whether to run grading in debug mode (prints grading STDOUT and STDERR 
            from each container to the command line)
        seed (``int``, optional): a random seed to use for intercell seeding during grading
        zips (``bool``, False): whether the submissions are zip files formatted from ``Notebook.export``

    Returns:
        ``pandas.core.frame.DataFrame``: A dataframe of file to grades information
    """
    # launch our docker conainer
    launch_command = ["docker", "run", "-d","-it", image]
    launch = subprocess.run(launch_command, stdout=PIPE, stderr=PIPE)
    
    # print(launch.stderr)
    container_id = launch.stdout.decode('utf-8')[:-1]

    if verbose:
        print(f"Launched container {container_id[:12]}...")
    
    # copy the notebook files to the container
    copy_command = ["docker", "cp", notebooks_dir, f"{container_id}:/home/notebooks/"]
    copy = subprocess.run(copy_command, stdout=PIPE, stderr=PIPE)
    
    # copy the test files to the container
    if tests_dir is not None:
        tests_command = ["docker", "cp", tests_dir, f"{container_id}:/home/tests/"]
        tests = subprocess.run(tests_command, stdout=PIPE, stderr=PIPE)

    # copy the requirements file to the container
    if reqs:
        if verbose:
            print(f"Installing requirements in container {container_id[:12]}...")
        reqs_command = ["docker", "cp", reqs, f"{container_id}:/home"]
        requirements = subprocess.run(reqs_command, stdout=PIPE, stderr=PIPE)

        # install requirements
        install_command = [
            "docker", "exec", "-t", container_id, "pip3", "install", "-r", "/home/requirements.txt"
        ]
        install = subprocess.run(install_command, stdout=PIPE, stderr=PIPE)

    if verbose:
        print(f"Grading {('notebooks', 'scripts')[scripts]} in container {container_id[:12]}...")
    
    # Now we have the notebooks in home/notebooks, we should tell the container to execute the 
    # grade command...
    grade_command = [
        "docker", "exec", "-t", container_id, "python3", "-m", "otter", "/home/notebooks"
    ]

    # if we want PDF output, add the necessary flag
    # if unfiltered_pdfs:
    #     grade_command += ["--pdf"]
    # elif tag_filter:
    #     grade_command += ["--pdf", "tags"]
    # elif html_filter:
    #     grade_command += ["--pdf", "html"]
    if pdfs:
        grade_command += ["--pdf", pdfs]
    
    # seed
    if seed is not None:
        grade_command += ["--seed", str(seed)]

    # if we are grading scripts, add the --script flag
    if scripts:
        grade_command += ["--scripts"]
    
    if debug:
        grade_command += ["--verbose"]

    if zips:
        grade_command += ["--zips"]

    grade = subprocess.run(grade_command, stdout=PIPE, stderr=PIPE)
    
    if debug:
        print(grade.stdout.decode("utf-8"))
        print(grade.stderr.decode("utf-8"))

    all_commands = [launch, copy, grade]
    try: 
        all_commands += [tests]
    except UnboundLocalError:
        pass
    try:
        all_commands += [requirements, install]
    except UnboundLocalError:
        pass

    try:
        for command in all_commands:
            if command.stderr.decode('utf-8') != '':
                raise Exception(f"Error running {command}; failed with error:\n{command.stderr.decode('utf-8')}")

        if verbose:
            print(f"Copying grades from container {container_id[:12]}...")

        # get the grades back from the container and read to date frame so we can merge later
        csv_command = [
            "docker", "cp", f"{container_id}:/home/notebooks/grades.csv", f"./grades{id}.csv"
        ]
        csv = subprocess.run(csv_command, stdout=PIPE, stderr=PIPE)
        df = pd.read_csv(f"./grades{id}.csv")

        if pdfs:
            pdf_folder = os.path.join(os.path.abspath(output_path), "submission_pdfs")
            os.makedirs(pdf_folder, exist_ok=True)
            
            # copy out manual submissions
            for pdf in df["manual"]:
                copy_cmd = [
                    "docker", "cp", f"{container_id}:{pdf}", os.path.join(pdf_folder, os.path.basename(pdf))
                ]
                copy = subprocess.run(copy_cmd, stdout=PIPE, stderr=PIPE)

            df["manual"] = df["manual"].str.replace("/home/notebooks", "submission_pdfs")
        
        # delete the file we just read
        csv_cleanup_command = ["rm", f"./grades{id}.csv"]
        csv_cleanup = subprocess.run(csv_cleanup_command, stdout=PIPE, stderr=PIPE)

        if not no_kill:
            if verbose:
                print(f"Stopping container {container_id[:12]}...")

            # cleanup the docker container
            stop_command = ["docker", "stop", container_id]
            stop = subprocess.run(stop_command, stdout=PIPE, stderr=PIPE)
            remove_command = ["docker", "rm", container_id]
            remove = subprocess.run(remove_command, stdout=PIPE, stderr=PIPE)

    except:
        # delete the file we just read
        csv_cleanup_command = ["rm", f"./grades{id}.csv"]
        csv_cleanup = subprocess.run(csv_cleanup_command, stdout=PIPE, stderr=PIPE)

        # delete the submission PDFs on failure
        if pdfs:
            csv_cleanup_command = ["rm", "-rf", os.path.join(output_path, "submission_pdfs")]
            csv_cleanup = subprocess.run(csv_cleanup_command, stdout=PIPE, stderr=PIPE)

        if not no_kill:
            if verbose:
                print(f"Stopping container {container_id[:12]}...")

            # cleanup the docker container
            stop_command = ["docker", "stop", container_id]
            stop = subprocess.run(stop_command, stdout=PIPE, stderr=PIPE)
            remove_command = ["docker", "rm", container_id]
            remove = subprocess.run(remove_command, stdout=PIPE, stderr=PIPE)
        
        raise
    
    # check that no commands errored, if they did rais an informative exception
    all_commands = [launch, copy, grade, csv, csv_cleanup]

    try:
        all_commands += [stop, remove]
    except UnboundLocalError:
        pass

    try:
        all_commands += [tests]
    except UnboundLocalError:
        pass

    try:
        all_commands += [requirements, install]
    except UnboundLocalError:
        pass

    for command in all_commands:
        for command in all_commands:
            if command.stderr.decode('utf-8') != '':
                raise Exception(f"Error running {command}; failed with error:\n{command.stderr.decode('utf-8')}")
    
    return df



################################
##### Tests for Otter Logs #####
################################

import unittest

# read in argument parser
bin_globals = {}

with open("bin/otter") as f:
    exec(f.read(), bin_globals)

parser = bin_globals["parser"]

TEST_FILES_PATH = "test/test-logs/"

class TestLogs(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print("\n\n\n" + ("=" * 60) + f"\nRunning {__name__}.{cls.__name__}\n" + ("=" * 60) + "\n")
    
    def test_something(self):
        pass



#############################################
##### Utilities for Otter on Gradescope #####
#############################################

import re
import nbformat

from ..assign.utils import get_source

def replace_notebook_instances(nb_path):
    """
    Replaces the creation of ``otter.Notebook`` instances in a notebook at ``nb_path`` that have custom 
    test paths with the default for Gradescope.

    Args:
        nb_path (``str``): path to the notebook
    """
    nb = nbformat.read(nb_path, as_version=nbformat.NO_CONVERT)

    instance_regex = r"otter.Notebook\([\"'].+[\"']\)"
    for cell in nb['cells']:
        source = get_source(cell)
        for i, line in enumerate(source):
            line = re.sub(instance_regex, "otter.Notebook()", line)
            source[i] = line
        cell['source'] = "\n".join(source)

    nbformat.write(nb, nb_path)



###########################################
##### OK Test Parser for Otter-Grader #####
#####     forked from Gofer-Grader    #####
###########################################

import ast
import os
import doctest
import io
import string
import warnings

from glob import glob
from contextlib import redirect_stderr, redirect_stdout
from textwrap import dedent
from jinja2 import Template
from pygments import highlight
from pygments.lexers import PythonConsoleLexer
from pygments.formatters import HtmlFormatter

from .utils import hide_outputs


class CheckCallWrapper(ast.NodeTransformer):
    """Visits and replaces nodes in an abstract syntax tree in-place.
    
    Tracks import syntax and instances of ``otter.Notebook`` in an AST. Wraps calls to 
    ``otter.Notebook.check`` in calls to ``list.append`` to collect results of execution. Removes calls
    to ``otter.Notebook.check_all``, `otter.Notebook.export``, and ``otter.Notebook.to_pdf``.
    
    Args:
        secret (``str``): random digits string that prevents check function from being modified
    
    Attributes:
        secret (``str``): random digits string that prevents check function from being modified
    """
    OTTER_IMPORT_SYNTAX = "import"
    OTTER_IMPORT_NAME = "otter"
    OTTER_CLASS_NAME = "Notebook"
    OTTER_INSTANCE_NAME = "grader"

    def __init__(self, secret):
        self.secret = secret

    def check_node_constructor(self, expression):
        """Creates node that wraps expression in a list (``check_results_XX``) append call
        
        Args:
            expression (``ast.Name``): name for check function

        Returns:
            ``ast.Call``: function call object from calling check

        """
        args = [expression]
        func = ast.Attribute(
            attr='append',
            value=ast.Name(id='check_results_{}'.format(self.secret), ctx=ast.Load()),
            ctx=ast.Load(),
            keywords=[]
        )
        return ast.Call(func=func, args=args, keywords=[])

    def visit_ImportFrom(self, node):
        """
        Visits ``from ... import ...`` nodes and tracks the import syntax and alias of ``otter.Notebook``

        Args:
            node (``ast.ImportFrom``): the import node

        Returns:
            ``ast.ImportFrom``: the original node
        """
        if node.module == "otter" and "Notebook" in [n.name for n in node.names]:
            CheckCallWrapper.OTTER_IMPORT_SYNTAX = "from"
            nb_asname = [n.asname for n in node.names if n.name == "Notebook"][0]
            if nb_asname is not None:
                CheckCallWrapper.OTTER_CLASS_NAME = nb_asname
        return node

    def visit_Import(self, node):
        """
        Visits ``import ...`` nodes and tracks the import syntax and alias of ``otter``

        Args:
            node (``ast.Import``): the import node

        Returns:
            ``ast.Import``: the original node
        """
        if "otter" in [n.name for n in node.names]:
            CheckCallWrapper.OTTER_IMPORT_SYNTAX = "import"
            otter_asname = [n.asname for n in node.names if n.name == "otter"][0]
            if otter_asname is not None:
                CheckCallWrapper.OTTER_IMPORT_NAME = otter_asname
        return node

    def visit_Assign(self, node):
        """
        Visits assignment nodes to determine the name assigned to the instance of ``otter.Notebook``
        created.

        Args:
            node (``ast.Assign``): the assignment node

        Returns:
            ``ast.Assign``: the original node
        """
        if isinstance(node.value, ast.Call):
            if isinstance(node.value.func, ast.Attribute) and CheckCallWrapper.OTTER_IMPORT_SYNTAX == "import":
                if node.value.func.attr == "Notebook" and isinstance(node.value.func.value, ast.Name):
                    if node.value.func.value.id == CheckCallWrapper.OTTER_IMPORT_NAME:
                        assert len(node.targets) == 1, "error parsing otter.Notebook instance in ast"
                        CheckCallWrapper.OTTER_INSTANCE_NAME = node.targets[0].id
            elif isinstance(node.value.func, ast.Name) and CheckCallWrapper.OTTER_IMPORT_SYNTAX == "from":
                if node.value.func.id == CheckCallWrapper.OTTER_CLASS_NAME:
                    assert len(node.targets) == 1, "error parsing otter.Notebook instance in ast"
                    CheckCallWrapper.OTTER_INSTANCE_NAME = node.targets[0].id
        return node

    def visit_Expr(self, node):
        """
        Visits expression nodes and removes them if they are calls to ``otter.Notebook.check_all``,
        ``otter.Notebook.export``, or ``otter.Notebook.to_pdf`` or wraps them using 
        ``CheckCallWrapper.check_node_constructor`` if they are calls to ``otter.Notebook.check``.

        Args:
            node (``ast.Expr``): the expression node

        Returns:
            ``ast.Expr``: the transformed node
            ``None``: if the node was a removed call
        """
        if isinstance(node.value, ast.Call):
            call_node = node.value
            if isinstance(call_node.func, ast.Attribute):
                if isinstance(call_node.func.value, ast.Name) and call_node.func.value.id == CheckCallWrapper.OTTER_INSTANCE_NAME:
                    if call_node.func.attr in ["check_all", "export", "to_pdf"]:
                        return None
                    elif call_node.func.attr == "check":
                        node.value = self.check_node_constructor(call_node)
        return node


def run_doctest(name, doctest_string, global_environment):
    """
    Run a single test with given ``global_environment``. Returns ``(True, '')`` if the doctest passes. 
    Returns ``(False, failure_message)`` if the doctest fails.

    Args:
        name (``str``): name of doctest
        doctest_string (``str``): doctest in string form
        global_environment (``dict``): global environment resulting from the execution of a python 
            script/notebook
    
    Returns:
        ``tuple`` of (``bool``, ``str``): results from running the test
    """
    examples = doctest.DocTestParser().parse(
        doctest_string,
        name
    )
    test = doctest.DocTest(
        [e for e in examples if isinstance(e, doctest.Example)],
        global_environment,
        name,
        None,
        None,
        doctest_string
    )

    doctestrunner = doctest.DocTestRunner(verbose=True)

    runresults = io.StringIO()
    with redirect_stdout(runresults), redirect_stderr(runresults), hide_outputs():
        doctestrunner.run(test, clear_globs=False)
    with open(os.devnull, 'w') as f, redirect_stderr(f), redirect_stdout(f):
        result = doctestrunner.summarize(verbose=True)
    # An individual test can only pass or fail
    if result.failed == 0:
        return (True, '')
    else:
        return False, runresults.getvalue()


class OKTest:
    """
    A single test (set of doctests) for Otter.
    
    Instances of this class are callable. When called, it takes a ``global_environment`` dict, and returns 
    an ``otter.ok_parser.OKTestsResult`` object. We only take a global_environment, *not* a 
    ``local_environment``. This makes tests not useful inside functions, methods or other scopes with 
    local variables. This is a limitation of doctest, so we roll with it.

    The last 2 attributes (``passed``, ``failed_test``) are set after calling ``run()``.

    Args:
        name (``str``): name of test
        tests (``list`` of ``str``): ;ist of docstring tests to be run
        hiddens (``list`` of ``bool``): visibility of each tests in ``tests``
        value (``int``, optional): point value of this test, defaults to 1
        hidden (``bool``, optional): wheter this test is hidden
    """
    html_result_pass_template = Template("""
    <p><strong>{{ name }}</strong> passed!</p>
    """)

    plain_result_pass_template = Template("{{ name }} passed!")

    html_result_fail_template = Template("""
    <p><strong style='color: red;'>{{ name }}</strong></p>
    <p><strong>Test code:</strong><pre>{{test_code}}</pre></p>
    <p><strong>Test result:</strong><pre>{{test_result}}</pre></p>
    """)

    plain_result_fail_template = Template(dedent("""\
       {{ name }}

    Test result:
    {{test_result}}"""))

    def _repr_html_(self):
        if self.passed:
            return OKTest.html_result_pass_template.render(name=self.name)
        else:
            return OKTest.html_result_fail_template.render(
                name=self.name,
                test_code=highlight(self.failed_test, PythonConsoleLexer(), HtmlFormatter(noclasses=True)),
                test_result=self.result
            )

    def __repr__(self):
        if self.passed:
            return OKTest.plain_result_pass_template.render(name=self.name)
        else:
            return OKTest.plain_result_fail_template.render(
                name=self.name,
                test_code=self.failed_test, # highlight(self.failed_test, PythonConsoleLexer(), HtmlFormatter(noclasses=True)),
                test_result=self.result
            )

    def __init__(self, name, tests, hiddens, value=1, hidden=True):
        self.name = name
        self.public_tests = [t for t, h in zip(tests, hiddens) if not h]
        self.hidden_tests = [t for t, h in zip(tests, hiddens) if h]
        self.value = value
        self.hidden = hidden
        self.passed = None
        self.failed_test = None
        self.failed_test_hidden = None

    def run(self, global_environment):
        """Runs tests on a given ``global_environment``
        
        Arguments:
            ``global_environment`` (``dict``): result of executing a Python notebook/script
        
        Returns:
            ``tuple`` of (``bool``, ``otter.ok_parser.OKTest``): whether the test passed and a pointer 
                to the current ``otter.ok_parser.OKTest`` object
        """
        for i, t in enumerate(self.public_tests + self.hidden_tests):
            passed, result = run_doctest(self.name + ' ' + str(i), t, global_environment)
            if not passed:
                self.passed = False
                self.failed_test = t
                self.failed_test_hidden = i >= len(self.public_tests)
                self.result = result
                return False, self
        self.passed = True
        return True, self

    @classmethod
    def from_file(cls, path):
        """
        Parse an ok test file & return an ``OKTest``

        Args:
            path (``str``): path to ok test file

        Returns:
            ``otter.ok_parser.OKTest``: new ``OKTest`` object created from the given file
        """
        # ok test files are python files, with a global 'test' defined
        test_globals = {}
        with open(path) as f:
            exec(f.read(), test_globals)

        test_spec = test_globals['test']

        # We only support a subset of these tests, so let's validate!

        # Make sure there is a name
        assert 'name' in test_spec

        # Do not support multiple suites in the same file
        assert len(test_spec['suites']) == 1

        test_suite = test_spec['suites'][0]

        # Only support doctest. I am unsure if other tests are implemented
        assert test_suite.get('type', 'doctest') == 'doctest'

        # Not setup and teardown supported
        assert not bool(test_suite.get('setup'))
        assert not bool(test_suite.get('teardown'))

        if 'hidden' in test_spec:
            warnings.warn(
                "The global 'hidden' key of ok-tests is deprecated since v1.0.0. "
                "This key will be ignored.", 
                FutureWarning
            )

        tests = []
        hiddens = []

        for _, test_case in enumerate(test_spec['suites'][0]['cases']):
            tests.append(dedent(test_case['code']))
            hiddens.append(test_case.get('hidden', True))

        return cls(path, tests, hiddens, test_spec.get('points', 1), test_spec.get('hidden', True))


class OKTests:
    """Test class for ok-style tests used to grade assignments
    
    Args:
        test_paths (``list`` of ``str``): list of paths to ok tests
    """
    def __init__(self, test_paths):
        self.paths = test_paths
        self.tests = [OKTest.from_file(path) for path in self.paths if "__init__.py" not in path]

    def run(self, global_environment, include_grade=True):
        """Run this object's tests on a given global environment (from running a notebook/script)
        
        Arguments:
            global_environment (``dict``): result of executing a Python notebook/script
            include_grade (``bool``, optional): whether grade should be included in result
        
        Returns:
            ``otter.ok_parser.OKTestsResult``: object resulting from running tests on ``global_environment`` 
                with grade, tests passed, and more information
        """
        passed_tests = []
        failed_tests = []
        grade = 0
        total = 0
        for t in self.tests:
            total += t.value
            passed, test_obj = t.run(global_environment)
            if passed:
                grade += t.value
                passed_tests.append(t)
            else:
                failed_tests.append((t, test_obj))

        try:
            grade /= total
        except ZeroDivisionError:
            grade = 0

        return OKTestsResult(grade, self.paths, self.tests, passed_tests,
                             failed_tests, include_grade)


class OKTestsResult:
    """
    Displayable result generated from ``otter.ok_parser.OKTests``

    Args:
        grade (``float``): grade as a decimal in the range [0, 1]
        paths (``list`` of ``str``): list of paths to ok tests
        tests (``list`` of ``OKTest``): list of OKTest objects for each path
        passed_tests (``list`` of ``str``): list of passed test docstrings
        failed_tests (``list`` of ``str``, ``otter.ok_parser.OKTest``): list of failed test docstrings and
            ``OKTest`` objects
        include_grade (``bool``, optional): whether grade should be included in result

    Attributes:
        grade (``float``): grade as a decimal in the range [0, 1]
        paths (``list`` of ``str``): list of paths to ok tests
        tests (``list`` of ``OKTest``): list of OKTest objects for each path
        passed_tests (``list`` of ``str``): list of passed test docstrings
        failed_tests (``list`` of ``str``, ``otter.ok_parser.OKTest``): list of failed test docstrings and
            ``OKTest`` objects
        include_grade (``bool``, optional): whether grade should be included in result

    """

    html_result_template = Template("""
    {% if include_grade %}
    <strong>Grade: {{ grade }}</strong>
    {% endif %}
    {% if grade == 1.0 %}
        <p>All tests passed!</p>
    {% else %}
        <p>{{ passed_tests|length }} of {{ tests|length }} tests passed</p>
        {% if passed_tests %}
        <p> <strong>Tests passed:</strong>
            {% for passed_test in passed_tests %} {{ passed_test.name }} {% endfor %}
        </p>
        {% endif %}
        {% if failed_tests %}
        <p> <strong>Tests failed: </strong>
            <ul>
            {% for failed_test, failed_test_obj in failed_tests %}
                <li> {{ failed_test_obj._repr_html_() }} </li>
            {% endfor %}
            </ul>
        {% endif %}
    {% endif %}
    """)

    plain_result_template = Template("""{% if include_grade %}Grade: {{ grade }}{% endif %}
    {% if grade == 1.0 %}All tests passed!{% else %}
    {{ passed_tests|length }} of {{ tests|length }} tests passed
    {% if passed_tests %}
    Tests passed: {% for passed_test in passed_tests %} {{ passed_test.name }} {% endfor %}{% endif %}
    {% if failed_tests %}
    Tests failed:
    {% for failed_test, failed_test_obj in failed_tests %}
        {{ failed_test_obj.__repr__() }}
    {% endfor %}
    {% endif %}
    {% endif %}
    """)

    def __init__(self, grade, paths, tests, passed_tests, failed_tests, include_grade=True):
        self.grade = grade
        self.paths = paths
        self.tests = tests
        self.passed_tests = passed_tests
        self.failed_tests = failed_tests
        self.include_grade = include_grade

    def _repr_html_(self):
        return OKTestsResult.html_result_template.render(
            grade=self.grade,
            passed_tests=self.passed_tests,
            failed_tests=self.failed_tests,
            tests=self.tests,
            include_grade=self.include_grade
        )

    def _repr_gradescope_(self):
        return OKTestsResult.plain_result_template.render(
            grade=self.grade,
            passed_tests=self.passed_tests,
            failed_tests=self.failed_tests,
            tests=self.tests,
            include_grade=self.include_grade
        )

    def __repr__(self):
        return OKTestsResult.plain_result_template.render(
            grade=self.grade,
            passed_tests=self.passed_tests,
            failed_tests=self.failed_tests,
            tests=self.tests,
            include_grade=self.include_grade
        )



#############################################
##### Create Database for Otter Service #####
#############################################

MISSING_PACKAGES = False

try:
    import yaml
    import csv
    import psycopg2
    import hashlib
    import os

    from psycopg2 import connect, extensions, sql
    from psycopg2.errors import DuplicateTable

    from ..utils import connect_db

except ImportError:
    # don't need requirements to use otter without otter service
    MISSING_PACKAGES = True

def create_users(filepath, host=None, username=None, password=None, conn=None):
    """
    Inserts usernames from filepath into db

    Args:
        filepath (str): Path to CSV files with usernames in first column, 
            passwords in second
        host (str, optional): Hostname where postgresql is running
        username (str, optional): Username for writing to postgres database
        password (str, optional): Password corresponding to provided username
    """
    with open(filepath, newline='') as csvfile:
        filereader = csv.reader(csvfile, delimiter=',', quotechar='|')
        # TODO: fill in the arguments below
        if conn is None:
            conn = connect_db(host, username, password)
        cursor = conn.cursor()
        for row in filereader:
            username, password = row[:2]
            password = hashlib.sha256(password.encode()).hexdigest()
            if username.lower() == "username":
                # skip heading
                continue
            insert_command = """
                INSERT INTO users (username, password) VALUES (\'{}\', \'{}\')
                ON CONFLICT (username)
                DO UPDATE SET password = \'{}\'
                """.format(username, password, password)
            cursor.execute(insert_command)

def remove_users(filepath, host=None, username=None, password=None, conn=None):
    """Removes users specified in file FILEPATH
    
    Args:
        filepath (str): Path to CSV files with usernames in first column, 
            passwords in second
        host (str, optional): Hostname where postgresql is running
        username (str, optional): Username for writing to postgres database
        password (str, optional): Password corresponding to provided username
    """
    with open(filepath, newline='') as csvfile:
        filereader = csv.reader(csvfile, delimiter=',', quotechar='|')
        # TODO: fill in the arguments below
        if conn is None:
            conn = connect_db(host, username, password)
        cursor = conn.cursor()
        remove_count = 0
        for row in filereader:
            username, password = row[:2]
            password = hashlib.sha256(password.encode()).hexdigest()
            if username.lower() == "username":
                # skip heading
                continue
            insert_command = """
                DELETE FROM users WHERE username = \'{}\'
                """.format(username)
            cursor.execute(insert_command)

def main(args, conn=None, close_conn=True):
    """
    conn is arg to pre-made db connection
    """
    if MISSING_PACKAGES:
        raise ImportError(
            "Missing some packages required for otter service. "
            "Please install all requirements at "
            "https://raw.githubusercontent.com/ucbds-infra/otter-grader/master/requirements.txt"
        )

    # # # TODO: can't assume we're in directory with conf.yml
    # # with open("conf.yml") as f:
    # #     config = yaml.safe_load(f)
    # try:
    #     assert os.path.isfile("conf.yml"), "conf.yml does not exist"
    #     with open("conf.yml") as f:
    #         config = yaml.safe_load(f)
    # except AssertionError:
    #     conf_path = input("What is the path of your config file?")
    #     with open(conf_path) as f:
    #         config = yaml.safe_load(f)
    pgconn = connect_db(args.db_host, args.db_user, args.db_pass, args.db_port, db='postgres')

    try:
        pgconn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        cursor = pgconn.cursor()
        cursor.execute('CREATE DATABASE otter_db')
    except psycopg2.errors.DuplicateDatabase:
        pass
    finally: 
        cursor.close()
        pgconn.close()

    if conn is None:
        conn = connect_db(args.db_host, args.db_user, args.db_pass, args.db_port)

    # conn = connect_db(args.db_host, args.db_user, args.db_pass, args.db_port)

    conn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    cursor = conn.cursor()

    queries = [
        '''
        CREATE TABLE users (
            user_id SERIAL PRIMARY KEY,
            api_keys VARCHAR[] CHECK (cardinality(api_keys) > 0),
            username TEXT UNIQUE,
            password VARCHAR,
            email TEXT UNIQUE,
            CONSTRAINT has_username_or_email CHECK (username IS NOT NULL or email IS NOT NULL)
        )
        ''',
        '''
        CREATE TABLE classes (
            class_id TEXT PRIMARY KEY,
            class_name TEXT NOT NULL
        )
        ''',
        '''
        CREATE TABLE assignments (
            assignment_id TEXT NOT NULL,
            class_id TEXT REFERENCES classes (class_id) NOT NULL,
            assignment_name TEXT NOT NULL,
            seed INTEGER,
            PRIMARY KEY (assignment_id, class_id)
        )
        ''',
        '''
        CREATE TABLE submissions (
            submission_id SERIAL PRIMARY KEY,
            assignment_id TEXT NOT NULL,
            class_id TEXT NOT NULL,
            user_id INTEGER REFERENCES users(user_id) NOT NULL,
            file_path TEXT NOT NULL,
            timestamp TIMESTAMP NOT NULL,
            score JSONB,
            FOREIGN KEY (assignment_id, class_id) REFERENCES assignments (assignment_id, class_id)
        )
        '''
    ]

    for query in queries:
        try:
            cursor.execute(query)
        except DuplicateTable:
            pass
        
    cursor.close()
    if close_conn:
        conn.close()



#################################################
##### Cointainer Building for Otter Service #####
#################################################

MISSING_PACKAGES = False

try:
    import subprocess
    import shutil
    import os
    import yaml

    from subprocess import PIPE, DEVNULL
    from io import BytesIO
    from jinja2 import Template
    from psycopg2.errors import UniqueViolation

    from ..utils import connect_db

except ImportError:
    MISSING_PACKAGES = True

DOCKERFILE_TEMPLATE = Template("""
FROM {{ image }}
RUN mkdir /home/notebooks
ADD {{ test_folder_path }} /home/tests{% if global_requirements %}
ADD {{ global_requirements }} /home
RUN pip3 install -r /home/{{ global_requirements_filename }}{% endif %}{% if requirements %}
ADD {{ requirements }} /home
RUN pip3 install -r /home/{{ requirements_filename }}{% endif %}{% for file in files %}
ADD {{ file }} /home/notebooks{% endfor %}
""")

def write_class_info(class_id, class_name, conn):
    """Writes the given class_name to the database, auto-generating a class_id

    Args:
        class_name (str): Name of new class to add to the database
        conn (Connection): Connection object for database
    
    Returns:
        class_id: Class ID for newly added class.
    """
    cursor = conn.cursor()
    insert_command = "INSERT INTO classes (class_id, class_name) \
        VALUES(%s, %s)"#.format(class_name)
    try:
        cursor.execute(insert_command, (class_id, class_name))
    except UniqueViolation:
        pass
    conn.commit()
    cursor.close()
    return class_id

def write_assignment_info(assignment_id, class_id, assignment_name, seed, conn):
    """Inserts/Updates assignment class_id and assignment_name in database. Inserts if assignment_id
        is not present. Updates if assignment_id is present.

    Args:
        assignment_id (str): Assignment id which will be created or updated
        class_id (str): Class id to insert/update
        assignment_name (str): Assignment name to insert/update
        conn (Connection): Connection object for database
    """
    cursor = conn.cursor()
    find_sql_command = """SELECT * FROM assignments
    WHERE assignment_id = '{}' AND class_id = '{}'
    """.format(assignment_id, class_id)
    cursor.execute(find_sql_command)

    # if seed is None:
    #     seed = "NULL"

    # If conflict on assignment_id, class_id
    # Update assignment_name
    if cursor.rowcount == 1:
        sql_command = """UPDATE assignments
                        SET assignment_name = %s, seed = %s
                        WHERE assignment_id = %s AND class_id = %s
                        """#.format(assignment_id, class_id, assignment_name, assignment_name)
        cursor.execute(sql_command, (assignment_name, seed, assignment_id, class_id))
    # Else, just insert
    else:
        sql_command = """INSERT INTO assignments (assignment_id, class_id, assignment_name, seed)
                        VALUES (%s, %s, %s, %s)
                        """#.format(assignment_id, class_id, assignment_name, assignment_name)
        cursor.execute(sql_command, (assignment_id, class_id, assignment_name, seed))

    conn.commit()
    cursor.close()

def main(args, conn=None, close_conn=True):
    if MISSING_PACKAGES:
        raise ImportError(
            "Missing some packages required for otter service. "
            "Please install all requirements at "
            "https://raw.githubusercontent.com/ucbds-infra/otter-grader/master/requirements.txt"
        )

    repo_path = args.repo_path
    assert os.path.exists(repo_path) and os.path.isdir(repo_path), "{} does not exist or is not a directory".format(repo_path)

    current_dir = os.getcwd()
    os.chdir(repo_path)

    # parse conf.yml
    assert os.path.isfile("conf.yml"), "conf.yml does not exist"
    with open("conf.yml") as f:
        config = yaml.safe_load(f)

    assignments = config["assignments"]
    assignment_cfs = [(a["name"], a["assignment_id"], a.get("seed", None)) for a in assignments]
    ids = [a["assignment_id"] for a in assignments]
    assert len(ids) == len(set(ids)), "Found non-unique assignment IDs in conf.yml"

    # Use one global connection for all db-related commands
    if conn is None:
        conn = connect_db(args.db_host, args.db_user, args.db_pass, args.db_port)
    class_id = write_class_info(config["class_id"], config["class_name"], conn)

    # write to the database
    for name, assignment_id, seed in assignment_cfs:
        write_assignment_info(assignment_id, class_id, name, seed, conn)

    # TODO: start building docker images
    for a in assignments:
        requirements = a["requirements"] if "requirements" in a else ""
        global_requirements = config["requirements"] if "requirements" in config else ""

        dockerfile = DOCKERFILE_TEMPLATE.render(
            image = args.image,
            test_folder_path = a["tests_path"],
            test_folder_name = os.path.split(a["tests_path"])[1],
            requirements = requirements,
            requirements_filename = os.path.split(requirements)[1],
            global_requirements = global_requirements,
            global_requirements_filename = os.path.split(global_requirements)[1],
            files = a.get("files", [])
        )

        print("Building Docker image {}".format(class_id + "-" + a["assignment_id"]))
        
        # Build the docker image
        echo_dockerfile = subprocess.Popen(["echo", dockerfile], stdout=PIPE)
        echo_dockerfile.wait()
        
        build_out = subprocess.Popen(
            ["docker", "build", "-f", "-", ".", "-t", class_id + "-" + a["assignment_id"]],
            stdin=echo_dockerfile.stdout, 
            stdout=DEVNULL if args.quiet else None
        )
        build_out.wait()
        
        echo_dockerfile.stdout.close()

        print("Built Docker image {}".format(class_id + "-" + a["assignment_id"]))
    
    if close_conn:
        conn.close()
    
    os.chdir(current_dir)



######################################
##### Utilities for Otter-Grader #####
######################################

MISSING_PACKAGES = False

import os
import sys
import pathlib
import random
import string
import time
import pandas as pd

from contextlib import contextmanager
from IPython import get_ipython

try:
    from psycopg2 import connect, extensions
except ImportError:
    MISSING_PACKAGES = True


@contextmanager
def block_print():
    """
    Context manager that disables printing to stdout.
    """
    sys.stdout = open(os.devnull, 'w')
    try:
        yield
    finally:
        try:
            sys.stdout.close()
        except:
            pass
        sys.stdout = sys.__stdout__

def list_files(path):
    """Returns a list of all non-hidden files in a directory
    
    Args:
        path (``str``): path to a directory
    
    Returns:
        ``list`` of ``str``: list of filenames in the given directory

    """
    return [file for file in os.listdir(path) if os.path.isfile(os.path.join(path, file)) and file[0] != "."]

def merge_csv(dataframes):
    """Merges dataframes along the vertical axis
    
    Args:
        dataframes (``list`` of ``pandas.core.frame.DataFrame``): list of dataframes with same columns
    
    Returns:
        ``pandas.core.frame.DataFrame``: A merged dataframe resulting from 'stacking' all input dataframes

    """
    final_dataframe = pd.concat(dataframes, axis=0, join='inner').sort_index()
    return final_dataframe

def connect_db(host="localhost", username="admin", password="", port="5432", db="otter_db"):
    """Connects to a specific Postgres database with provided parameters/credentials

    Arguments:
        host (``str``, optional): hostname for database (default 'localhost')
        username (``str``, optional): username with proper read/write permissions for Postgres
        password (``str``, optional): password for provided username
        port (``str``, optional): port on which Postgres is running

    Returns:
        ``connection``: connection object for executing SQL commands on Postgres database

    Raises:
        ``ImportError``: if psycopg2 is not installed
    """
    if MISSING_PACKAGES:
        raise ImportError(
            "Missing some packages required for otter service. "
            "Please install all requirements at "
            "https://raw.githubusercontent.com/ucbds-infra/otter-grader/master/requirements.txt"
        )
        
    conn = connect(dbname=db,
               user=username,
               host=host,
               password=password,
               port=port)
    conn.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    return conn

def flush_inline_matplotlib_plots():
    """
    Flush matplotlib plots immediately, rather than asynchronously
    
    Basically, the inline backend only shows the plot after the entire cell executes, which means we 
    can't easily use a context manager to suppress displaying it. See https://github.com/jupyter-widgets/ipywidgets/issues/1181/ 
    and https://github.com/ipython/ipython/issues/10376 for more details. This function displays flushes 
    any pending matplotlib plots if we are using the inline backend. Stolen from 
    https://github.com/jupyter-widgets/ipywidgets/blob/4cc15e66d5e9e69dac8fc20d1eb1d7db825d7aa2/ipywidgets/widgets/interaction.py#L35
    """
    if 'matplotlib' not in sys.modules:
        # matplotlib hasn't been imported, nothing to do.
        return

    try:
        import matplotlib as mpl
        from ipykernel.pylab.backend_inline import flush_figures
    except ImportError:
        return
    # except KeyError:
    #     return

    if mpl.get_backend() == 'module://ipykernel.pylab.backend_inline':
        flush_figures()

@contextmanager
def hide_outputs():
    """
    Context manager for hiding outputs from ``display()`` calls. IPython handles matplotlib outputs 
    specially, so those are supressed too.
    """
    ipy = get_ipython()
    if ipy is None:
        # Not running inside ipython!
        yield
        return
    old_formatters = ipy.display_formatter.formatters
    ipy.display_formatter.formatters = {}
    try:
        yield
    finally:
        # flush_inline_matplotlib_plots()
        ipy.display_formatter.formatters = old_formatters

def id_generator(size=6, chars=string.ascii_uppercase + string.digits):
    """Used to generate a dynamic variable name for grading functions

    This function generates a random name using the given length and character set.
    
    Args:
        size (``int``): length of output name
        chars (``str``, optional): set of characters used to create function name
    
    Returns:
        ``str``: randomized string name for grading function
    """
    return ''.join(random.choice(chars) for _ in range(size))

# def str_to_doctest(code_lines, lines):
#     """
#     Converts a list of lines of Python code ``code_lines`` to a list of doctest-formatted lines ``lines``

#     Args:
#         code_lines (``list``): list of lines of python code
#         lines (``list``): set of characters used to create function name
    
#     Returns:
#         ``list`` of ``str``: doctest formatted list of lines
#     """
#     if len(code_lines) == 0:
#         return lines
#     line = code_lines.pop(0)
#     if line.startswith(" ") or line.startswith("\t"):
#         return str_to_doctest(code_lines, lines + ["... " + line])
#     elif line.startswith("except:") or line.startswith("elif ") or line.startswith("else:") or line.startswith("finally:"):
#         return str_to_doctest(code_lines, lines + ["... " + line])
#     elif len(lines) > 0 and lines[-1].strip().endswith("\\"):
#         return str_to_doctest(code_lines, lines + ["... " + line])
#     else:
#         return str_to_doctest(code_lines, lines + [">>> " + line])

def get_variable_type(obj):
    """Returns the fully-qualified type string of an object ``obj``

    Args:
        obj (object): the object in question

    Returns:
        ``str``: the fully-qualified type string
    """
    return type(obj).__module__ + "." + type(obj).__name__

def get_relpath(src, dst):
    """
    Returns the relative path from ``src`` to ``dst``

    Args:
        src (``pathlib.Path``): the source directory
        dst (``pathlib.Path``): the destination directory

    Returns:
        ``pathlib.Path``: the relative path
    """
    osrc = src
    ups = 0
    while True:
        try:
            dst.relative_to(src)
            break
        except ValueError:
            src = src.parent
            ups += 1
    return pathlib.Path(("../" * ups) / dst.relative_to(src))

def wait_for_save(filename, timeout=5):
    """
    Waits for `filename` to update, waiting up to `timeout` seconds. Returns True if a save was detected, 
    and False otherwise. Stolen from 
    https://github.com/okpy/ok-client/blob/25c9e85caf60c3fb68bd072419ee704da07c5764/client/api/notebook.py#L158

    Args:
        filename (``str``): the filename
        timeout (``int``, optional): number of seconds to wait
    
    Returns:
        ``bool``: whether the file updated
    """
    modification_time = os.path.getmtime(filename)
    start_time = time.time()
    while time.time() < start_time + timeout:
        if (os.path.getmtime(filename) > modification_time and
            os.path.getsize(filename) > 0):
            return True
        time.sleep(0.2)
    return False



########################################################
##### Notebook and File Execution for Otter-Grader #####
########################################################

import os
import re
import json
import ast
import math
import itertools
import inspect

from collections import namedtuple
from unittest import mock
from contextlib import redirect_stdout, redirect_stderr
from IPython import get_ipython
from IPython.display import display
try:
    from IPython.core.inputsplitter import IPythonInputSplitter
except ImportError:
    raise ImportError('IPython needs to be installed for notebook grading')

from .ok_parser import OKTests, CheckCallWrapper
from .utils import hide_outputs, id_generator

TestResult = namedtuple("TestResult", ["name", "score", "possible", "test", "hidden", "incorrect"])

class GradingResults:
    """
    Stores and wrangles test result objects. Initialize with a list of ``otter.ok_parser.OKTestsResult``
    objects and this class will store the results as named tuples so that they can be accessed/manipulated
    easily. Also contains methods to put the results into a nice ``dict`` format or into the correct
    format for Gradescope.

    Args:
        results (``list`` of ``otter.ok_parser.OKTestsResult``): the list of grading results
    
    Attributes:
        raw_results (``list`` of ``otter.ok_parser.OKTestsResult``): the results passed to the constructor
        results (``dict``): maps test names to ``TestResult`` named tuples containing the test result
            information
        total (numeric): the total points earned by the submission
        possible (numeric): the total points possible based on the tests
        tests (``list`` of ``str``): list of test names according to the keys of ``results``
    """
    def __init__(self, results):
        self.raw_results = results
        self.results = {}
        
        total_score, points_possible = 0, 0
        for result in results:
            for test in result.tests:
                test_name = os.path.splitext(os.path.basename(test.name))[0]
                tr = TestResult(
                    name = test_name,
                    score = result.grade * test.value,
                    possible = test.value,
                    test = test,
                    hidden = False,
                    incorrect = False
                )
                self.results[test_name] = tr

                total_score += result.grade * test.value
                points_possible += test.value

            for test, test_obj in result.failed_tests:
                test_name = os.path.splitext(os.path.basename(test.name))[0]
                if test_name in self.results:
                    self.results[test_name] = self.results[test_name]._replace(
                        hidden = test_obj.failed_test_hidden,
                        incorrect = True
                    )
                else:
                    self.results[test_name] = TestResult(
                        name = test_name,
                        score = 0,
                        possible = test.value,
                        test = test,
                        hidden = test_obj.failed_test_hidden,
                        incorrect = True
                    )
        
        self.total = total_score
        self.possible = points_possible
    
    @property
    def tests(self):
        return list(self.results.keys())
    
    def get_result(self, test_name):
        """
        Returns the ``TestResult`` named tuple corresponding to the test with name ``test_name``

        Args:
            test_name (``str``): the name of the desired test
        
        Returns:
            ``TestResult``: the results of that test
        """
        return self.results[test_name]

    def to_dict(self):
        """
        Converts these results into a dictinary, extending the fields of the named tuples in ``results``
        into key, value pairs in a ``dict``.

        Returns:
            ``dict``: the results in dictionary form
        """
        output = {}
        for test_name in self.tests:
            result = self.get_result(test_name)
            output[test_name] = dict(result._asdict())

        return output

    def to_gradescope_dict(self, config={}):
        """
        Converts these results into a dictionary formatted for Gradescope's autograder. Requires a 
        dictionary of configurations for the Gradescope assignment generated using Otter Generate.

        Args:
            config (``dict``): the grading configurations

        Returns:
            ``dict``: the results formatted for Gradescope
        """
        output = {"tests": []}

        # hidden visibility determined by show_hidden_tests_on_release
        hidden_test_visibility = ("hidden", "after_published")[config.get("show_hidden_tests_on_release", False)]
        no_separate_visibility = config.get("test_visibility", "hidden")
        assert no_separate_visibility in ["hidden", "visible", "after_published"]

        for test_name in self.tests:
            result = self.get_result(test_name)
            hidden, incorrect = result.hidden, result.incorrect
            score, possible = result.score, result.possible
            public_score, hidden_score = score * config.get("public_multiplier", 0), score * (1 - config.get("public_multiplier", 0))
            public_possible, hidden_possible = possible * config.get("public_multiplier", 0), possible * (1 - config.get("public_multiplier", 0))
        
            if hidden and incorrect:
                public_score, hidden_score = possible * config.get("public_multiplier", 0), 0
            elif not hidden and incorrect:
                public_score, hidden_score = 0, 0
                public_possible = possible
            
            if config.get("separate_tests", True):
                output["tests"].append({
                    "name": result.name + " - Public",
                    "score": public_score,
                    "max_score": public_possible,
                    "visibility": "visible",
                    "output": repr(result.test) if not hidden and incorrect else "All tests passed!"
                })

                if not (not hidden and incorrect):
                    output["tests"].append({
                        "name" : result.name + " - Hidden",
                        "score" : hidden_score,
                        "max_score": hidden_possible,
                        "visibility": hidden_test_visibility,
                        "output": repr(result.test) if incorrect else "All tests passed!"
                    })
            
            else:
                output["tests"].append({
                    "name": result.name,
                    "score": score,
                    "max_score": possible,
                    "visibility": no_separate_visibility,
                    "output": repr(result.test) if not hidden and incorrect else "All tests passed!"
                })
        
        if config.get("show_stdout_on_release", False):
            output["stdout_visibility"] = "after_published"

        if config.get("points_possible", None) is not None:
            output["score"] = self.total / self.possible * config.get("points_possible", None)

        if config.get("score_threshold", None) is not None:
            if self.total / self.possible >= config["score_threshold"]:
                output["score"] = config.get("points_possible", None) or self.possible
            else:
                output["score"] = 0
        
        return output

def check(test_file_path, global_env=None):
    """
    Checks ``global_env`` against given ``test_file`` in OK-format. If global_env is ``None``, the 
    global environment of the calling function is used. The following two calls are equivalent:

    .. code-block:: python
        check('tests/q1.py')
        check('tests/q1.py', globals())
    
    Args:
        test_file_path (``str``): path to test file
        global_env (``dict``, optional): a global environment resulting from the execution 
            of a python script or notebook

    Returns:
        ``otter.ok_parser.OKTestsResult``: result of running the tests in the given global environment

    """
    tests = OKTests([test_file_path])

    if global_env is None:
        # Get the global env of our callers - one level below us in the stack
        # The grade method should only be called directly from user / notebook
        # code. If some other method is calling it, it should also use the
        # inspect trick to pass in its parents' global env.
        global_env = inspect.currentframe().f_back.f_globals

    return tests.run(global_env, include_grade=False)

def grade_notebook(notebook_path, tests_glob=None, name=None, ignore_errors=True, script=False, 
    cwd=None, test_dir=None, seed=None, log=None, variables=None):
    """
    Grade a notebook file & return grade information

    This function grades a single Jupyter notebook using the provided tests. If grading a Python file,
    set ``script`` to true. 

    Args:
        notebook_path (``str``): path to a single notebook
        tests_glob (``list`` of ``str``, optional): names of test files
        name (``str``, optional): initial environment name
        ignore_errors (``bool``, optional): whether errors in execution should be ignored
        script (``bool``, optional): whether the notebook_path is a Python script
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        seed (``int``, optional): random seed for intercell seeding
        log (``otter.logs.Log``, optional): log from which to grade questions
        variables (``dict``, optional): map of variable names -> type string to check type of deserialized
            object to prevent arbitrary code from being put into the environment; ignored if log is ``None``

    Returns:
        ``GradingResults``: the results of grading
    """
    # ensure this is not being executed inside a notebook
    assert get_ipython() is None, "Cannot execute inside Jupyter Notebook"

    if not script:
        try:
            with open(notebook_path) as f:
                nb = json.load(f)
        except UnicodeDecodeError:
            with open(notebook_path, encoding='utf-8') as f:
                nb = json.load(f)
    else:
        with open(notebook_path) as f:
            nb = f.read()

    secret = id_generator()
    results_array = "check_results_{}".format(secret)
    initial_env = {
        results_array: []
    }

    if name:
        initial_env["__name__"] = name

    if log is not None:
        global_env = execute_log(nb, log, secret, initial_env, ignore_errors=ignore_errors, cwd=cwd, test_dir=test_dir, variables=variables)
    elif script:
        global_env = execute_script(nb, secret, initial_env, ignore_errors=ignore_errors, cwd=cwd, seed=seed)
    else:
        global_env = execute_notebook(nb, secret, initial_env, ignore_errors=ignore_errors, cwd=cwd, test_dir=test_dir, seed=seed)

    test_results = global_env[results_array]

    # Check for tests which were not included in the notebook and specified by tests_globs
    # Allows instructors to run notebooks with additional tests not accessible to user
    if tests_glob:
        # unpack list of paths into a single list
        tested_set = list(itertools.chain(*[r.paths for r in test_results]))
        extra_tests = []
        for t in sorted(tests_glob):
            include = True
            for tested in tested_set:
                if tested in t:     # e.g. if 'tests/q1.py' is in /srv/repo/lab01/tests/q1.py'
                    include = False
            if include:
                extra_tests.append(OKTests([t]))
        extra_results = [t.run(global_env, include_grade=False) for t in extra_tests]
        test_results += extra_results

    # score_mapping = {}
    # points_possible, total_score = 0, 0
    # for r in test_results:
    #     try:
    #         for test in r.tests:
    #             test_name = os.path.split(test.name)[1][:-3]
    #             score_mapping[test_name] = {
    #                 "score": r.grade * test.value,
    #                 "possible": test.value,
    #                 "test": test,
    #                 # "hidden": test.hidden
    #             }
    #             total_score += r.grade * test.value
    #             points_possible += test.value
    #         for tup in r.failed_tests:
    #             test_name = os.path.split(tup[0].name)[1][:-3]
    #             if test_name in score_mapping:
    #                 score_mapping[test_name]["hint"] = tup[1]#.__repr__()
    #                 score_mapping[test_name]["hidden"] = tup[1].failed_test_hidden
    #             else:
    #                 score_mapping[test_name] = {
    #                     "hint": tup[1],#.__repr__()
    #                     "hidden": tup[1].failed_test_hidden,
    #                     "test": tup[0],
    #                 }
    #     except IndexError:
    #         pass

    # # add in total score and avoid divide by zero error if there are no tests
    # score_mapping["total"] = total_score
    # score_mapping["possible"] = points_possible

    # return score_mapping

    return GradingResults(test_results)

def execute_log(nb, log, secret='secret', initial_env=None, ignore_errors=False, cwd=None, test_dir=None, variables=None):
    """
    Executes a notebook from logged environments and returns the global environment that results from execution

    Execute notebook & return the global environment that results from execution. If ``ignore_errors`` 
    is ``True``, exceptions are swallowed. ``secret`` contains random digits so ``check_results`` and 
    ``check`` are not easily modifiable. ``nb`` is passed in as a dictionary that's a parsed notebook

    Args:
        nb (``dict``): JSON representation of a notebook
        log (``otter.logs.Log``): log from notebook execution
        secret (``str``, optional): randomly generated integer used to rebind check function
        initial_env (``str``, optional): name of initial environment
        ignore_errors (``bool``, optional): whether exceptions should be ignored
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        variables (``dict``, optional): map of variable names -> type string to check type of deserialized
            object to prevent arbitrary code from being put into the environment
    
    Results:
        ``dict``: global environment resulting from executing all code of the input notebook
    """
    with hide_outputs():
        if initial_env:
            global_env = initial_env.copy()
        else:
            global_env = {}

        if test_dir:
            source = f"import otter\ngrader = otter.Notebook(\"{test_dir}\")\n"
        else:
            source = f"import otter\ngrader = otter.Notebook()\n"
        
        if cwd:
            source +=  f"import sys\nsys.path.append(\"{cwd}\")\n"

        logged_questions = []
        m = mock.mock_open()
        with mock.patch("otter.notebook.Notebook._log_event", m):
            exec(source, global_env)

            for cell in nb['cells']:
                if cell['cell_type'] == 'code':
                    # transform the input to executable Python
                    # FIXME: use appropriate IPython functions here
                    isp = IPythonInputSplitter(line_input_checker=False)
                    
                    code_lines = []
                    cell_source_lines = cell['source']
                    source_is_str_bool = False
                    if isinstance(cell_source_lines, str):
                        source_is_str_bool = True
                        cell_source_lines = cell_source_lines.split('\n')

                    # only execute import statements
                    cell_source_lines = [re.sub(r"^\s+", "", l) for l in cell_source_lines if "import" in l]                                
                    
                    for line in cell_source_lines:
                        try:
                            exec(line, global_env)
                    # source += cell_source
                        except:
                            if not ignore_errors:
                                raise


            for entry in log.question_iterator():
                shelf = entry.unshelve(global_env)

                if variables is not None:
                    for k, v in shelf.items():
                        full_type = type(v).__module__ + "." + type(v).__name__
                        if not (k in variables and variables[k] == full_type):
                            del shelf[k]
                            print(f"Found variable of different type than expected: {k}")

                global_env.update(shelf)
                global_env[f"check_results_{secret}"].append(global_env["grader"].check(entry.question, global_env=global_env))
                logged_questions.append(entry.question)

        print("Questions executed from log: {}".format(", ".join(logged_questions)))
        
        return global_env

def execute_notebook(nb, secret='secret', initial_env=None, ignore_errors=False, cwd=None, test_dir=None, seed=None):
    """
    Executes a notebook and returns the global environment that results from execution

    Execute notebook & return the global environment that results from execution. If ``ignore_errors`` 
    is ``True``, exceptions are swallowed. ``secret`` contains random digits so ``check_results`` and 
    ``check`` are not easily modifiable. ``nb`` is passed in as a dictionary that's a parsed notebook

    Args:
        nb (``dict``): JSON representation of a notebook
        secret (``str``, optional): randomly generated integer used to rebind check function
        initial_env (``str``, optional): name of initial environment
        ignore_errors (``bool``, optional): whether exceptions should be ignored
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        seed (``int``, optional): random seed for intercell seeding
    
    Results:
        ``dict``: global environment resulting from executing all code of the input notebook
    """
    with hide_outputs():
        if initial_env:
            global_env = initial_env.copy()
        else:
            global_env = {}

        # add display from IPython
        global_env["display"] = display

        source = ""
        # if gradescope:
        #     source = "import sys\nsys.path.append(\"/autograder/submission\")\n"
        # el
        if cwd:
            source = f"import sys\nsys.path.append(\"{cwd}\")\n"
            exec(source, global_env)
        if seed is not None:
            # source += "import numpy as np\nimport random\n"
            import numpy as np
            import random
            global_env["np"] = np
            global_env["random"] = random

        # Before rewriting AST, find cells of code that generate errors.
        # One round of execution is done beforehand to mimic the Jupyter notebook style of running
        # (e.g. code runs up to the point of execution).
        # The reason this is workaround is introduced is because once the
        # source code is parsed into an AST, there is no sense of local cells

        for cell in nb['cells']:
            if cell['cell_type'] == 'code':
                # transform the input to executable Python
                # FIXME: use appropriate IPython functions here
                isp = IPythonInputSplitter(line_input_checker=False)
                try:
                    code_lines = []
                    cell_source_lines = cell['source']
                    source_is_str_bool = False
                    if isinstance(cell_source_lines, str):
                        source_is_str_bool = True
                        cell_source_lines = cell_source_lines.split('\n')

                    for line in cell_source_lines:
                        # Filter out ipython magic commands
                        # Filter out interact widget
                        if not line.startswith('%'):
                            if "interact(" not in line and not re.search(r"otter\.Notebook\(.*?\)", line):
                                code_lines.append(line)
                                if source_is_str_bool:
                                    code_lines.append('\n')
                            elif re.search(r"otter\.Notebook\(.*?\)", line):
                                # TODO: move this check into CheckCallWrapper
                                # if gradescope:
                                #     line = re.sub(r"otter\.Notebook\(.*?\)", "otter.Notebook(\"/autograder/submission/tests\")", line)
                                # el
                                if test_dir:
                                    line = re.sub(r"otter\.Notebook\(.*?\)", f"otter.Notebook(\"{test_dir}\")", line)
                                else:
                                    line = re.sub(r"otter\.Notebook\(.*?\)", "otter.Notebook(\"/home/tests\")", line)
                                code_lines.append(line)
                                if source_is_str_bool:
                                    code_lines.append('\n')
                    if seed is not None:
                        cell_source = "np.random.seed({})\nrandom.seed({})\n".format(seed, seed) + isp.transform_cell(''.join(code_lines))
                    else:
                        cell_source = isp.transform_cell(''.join(code_lines))

                    # patch otter.Notebook.export so that we don't create PDFs in notebooks
                    # TODO: move this patch into CheckCallWrapper
                    m = mock.mock_open()
                    with mock.patch('otter.Notebook.export', m), mock.patch("otter.Notebook._log_event", m):
                        exec(cell_source, global_env)
                    source += cell_source
                except:
                    if not ignore_errors:
                        raise

        tree = ast.parse(source)
        # # CODE BELOW COMMENTED OUT BECAUSE the only check function is within the Notebook class
        # if find_check_assignment(tree) or find_check_definition(tree):
        #     # an empty global_env will fail all the tests
        #     return global_env

        # wrap check(..) calls into a check_results_X.append(check(..))
        transformer = CheckCallWrapper(secret)
        tree = transformer.visit(tree)
        ast.fix_missing_locations(tree)

        cleaned_source = compile(tree, filename="nb-ast", mode="exec")
        try:
            with open(os.devnull, 'w') as f, redirect_stdout(f), redirect_stderr(f):
                # patch otter.Notebook.export so that we don't create PDFs in notebooks
                m = mock.mock_open()
                with mock.patch('otter.Notebook.export', m), mock.patch("otter.Notebook._log_event", m):
                    exec(cleaned_source, global_env)
        except:
            if not ignore_errors:
                raise
        return global_env

def execute_script(script, secret='secret', initial_env=None, ignore_errors=False, cwd=None, test_dir=None, seed=None):
    """
    Executes code of a Python script and returns the resulting global environment

    Execute script & return the global environment that results from execution. If ``ignore_errors`` is 
    ``True``, exceptions are swallowed. ``secret`` contains random digits so ``check_results`` and 
    ``check`` are not easily modifiable. ``script`` is passed in as a string.

    Args:
        script (``str``): string representation of Python script code
        secret (``str``, optional): randomly generated integer used to rebind check function
        initial_env (``str``, optional): name of initial environment
        ignore_errors (``bool``, optional): whether exceptions should be ignored
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        seed (``int``, optional): random seed for intercell seeding
    
    Results:
        dict: global environment resulting from executing all code of the input script
    """
    with hide_outputs():
        if initial_env:
            global_env = initial_env.copy()
        else:
            global_env = {}
        source = ""
        # if gradescope:
        #     source = "import sys\nsys.path.append(\"/autograder/submission\")\n"
        # el
        if cwd:
            source =  f"import sys\nsys.path.append(\"{cwd}\")\n"
        if seed is not None:
            # source += "import numpy as np\nimport random\n"
            import numpy as np
            import random
            global_env["np"] = np
            global_env["random"] = random
            source += "np.random.seed({})\nrandom.seed({})\n".format(seed, seed)

        lines = script.split("\n")
        for i, line in enumerate(lines):
            # TODO: move this check into CheckCallWrapper
            if re.search(r"otter\.Notebook\(.*?\)", line):
                # if gradescope:
                #     line = re.sub(r"otter\.Notebook\(.*?\)", "otter.Notebook(\"/autograder/submission/tests\")", line)
                # else:
                if test_dir:
                    line = re.sub(r"otter\.Notebook\(.*?\)", f"otter.Notebook(\"{test_dir}\")", line)
                else:
                    line = re.sub(r"otter\.Notebook\(.*?\)", "otter.Notebook(\"/home/tests\")", line)
            lines[i] = line
        try:
            exec("\n".join(lines), global_env)
            source += "\n".join(lines)
        except:
            if not ignore_errors:
                raise
        
        tree = ast.parse(source)
        # # CODE BELOW COMMENTED OUT BECAUSE the only check function is within the Notebook class
        # if find_check_assignment(tree) or find_check_definition(tree):
        #     # an empty global_env will fail all the tests
        #     return global_env

        # wrap check(..) calls into a check_results_X.append(check(..))
        transformer = CheckCallWrapper(secret)
        tree = transformer.visit(tree)
        ast.fix_missing_locations(tree)
        cleaned_source = compile(tree, filename="nb-ast", mode="exec")
        try:
            with open(os.devnull, 'w') as f, redirect_stdout(f), redirect_stderr(f):
                exec(cleaned_source, global_env)
        except:
            if not ignore_errors:
                raise
        return global_env



########################################
##### Run Autograder on Gradescope #####
########################################

import json
import os
import shutil
import subprocess
import re
import pickle
import warnings
import pandas as pd

from glob import glob
from textwrap import dedent

from .token import APIClient
from .utils import replace_notebook_instances

from ..execute import grade_notebook
from ..export import export_notebook
from ..logs import Log, QuestionNotInLogException
from ..notebook import _OTTER_LOG_FILENAME
from ..version import LOGO_WITH_VERSION


def main(config):
    """
    Runs autograder on Gradescope based on predefined configurations.

    Args:
        config (``dict``): configurations for autograder
    """
    print(LOGO_WITH_VERSION, "\n")

    if config.get("token", None) is not None:
        client = APIClient(token=config.get("token", None))
        generate_pdf = True
    else:
        generate_pdf = False

    # put files into submission directory
    if os.path.exists("/autograder/source/files"):
        for file in os.listdir("/autograder/source/files"):
            fp = os.path.join("/autograder/source/files", file)
            if os.path.isdir(fp):
                if not os.path.exists(os.path.join("/autograder/submission", os.path.basename(fp))):
                    shutil.copytree(fp, os.path.join("/autograder/submission", os.path.basename(fp)))
            else:
                shutil.copy(fp, "/autograder/submission")

    # create __init__.py files
    subprocess.run(["touch", "/autograder/__init__.py"])
    subprocess.run(["touch", "/autograder/submission/__init__.py"])

    os.chdir("/autograder/submission")

    # # check for *.ipynb.json files
    # jsons = glob("*.ipynb.json")
    # for file in jsons:
    #     shutil.copy(file, file[:-5])

    # # check for *.ipynb.html files
    # htmls = glob("*.ipynb.html")
    # for file in htmls:
    #     shutil.copy(file, file[:-5])

    nb_path = glob("*.ipynb")[0]

    replace_notebook_instances(nb_path)

    # # fix utils import
    # try:
    #     with open(nb_path) as f:
    #         contents = f.read()
    # except UnicodeDecodeError:
    #     with open(nb_path, "r", encoding="utf-8") as f:
    #         contents = f.read()
    
    # contents = re.sub(NOTEBOOK_INSTANCE_REGEX, "otter.Notebook()", contents)

    # try:
    #     with open(nb_path, "w") as f:
    #         f.write(contents)
    # except UnicodeEncodeError:
    #     with open(nb_path, "w", encoding="utf-8") as f:
    #         f.write(contents)

    os.makedirs("/autograder/submission/tests", exist_ok=True)
    tests_glob = glob("/autograder/source/tests/*.py")
    for file in tests_glob:
        shutil.copy(file, "/autograder/submission/tests")

    if glob("*.otter"):
        assert len(glob("*.otter")) == 1, "Too many .otter files (max 1 allowed)"
        with open(glob("*.otter")[0]) as f:
            otter_config = json.load(f)
    else:
        otter_config = None

    if os.path.isfile(_OTTER_LOG_FILENAME):
        log = Log.from_file(_OTTER_LOG_FILENAME, ascending=False)
        if config.get("grade_from_log", False):
            print("\n\n")     # for logging in otter.execute.execute_log
    else:
        assert not config.get("grade_from_log", False), "missing log"
        log = None

    scores = grade_notebook(
        nb_path, 
        glob("/autograder/submission/tests/*.py"), 
        name="submission", 
        cwd="/autograder/submission", 
        test_dir="/autograder/submission/tests",
        ignore_errors=not config.get("debug", False), 
        seed=config.get("seed", None),
        log=log if config.get("grade_from_log", False) else None,
        variables=config.get("serialized_variables", None)
    )

    # verify the scores against the log
    print("\n\n")
    if log is not None:
        try:
            found_discrepancy = log.verify_scores(scores)
            if not found_discrepancy:
                print("No discrepancies found while verifying scores against the log.")
        except BaseException as e:
            print(f"Error encountered while trying to verify scores with log:\n{e}")
    else:
        print("No log found with which to verify student scores")

    if generate_pdf:
        try:
            export_notebook(nb_path, filtering=config.get("filtering"), pagebreaks=config.get("pagebreaks"))
            pdf_path = os.path.splitext(nb_path)[0] + ".pdf"

            # get student email
            with open("../submission_metadata.json") as f:
                metadata = json.load(f)

            student_emails = []
            for user in metadata["users"]:
                student_emails.append(user["email"])
            
            for student_email in student_emails:
                client.upload_pdf_submission(config["course_id"], config["assignment_id"], student_email, pdf_path)

            print("\n\nSuccessfully uploaded submissions for: {}".format(", ".join(student_emails)))

        except:
            print("\n\n")
            warnings.warn("PDF generation or submission failed", RuntimeWarning)

    output = scores.to_gradescope_dict(config)

    # # hidden visibility determined by show_hidden_tests_on_release
    # hidden_test_visibility = ("hidden", "after_published")[config.get("show_hidden_tests_on_release", False)]

    # output = {"tests" : []}
    # for key in scores:
    #     if key != "total" and key != "possible":
    #         hidden, incorrect = scores[key].get("hidden", False), "hint" in scores[key]
    #         score, possible = scores[key]["score"], scores[key]["possible"]
    #         public_score, hidden_score = score * config.get("public_multiplier", 0), score * (1 - config.get("public_multiplier", 0))
    #         public_possible, hidden_possible = possible * config.get("public_multiplier", 0), possible * (1 - config.get("public_multiplier", 0))

    #         if hidden and incorrect:
    #             public_score, hidden_score = possible * config.get("public_multiplier", 0), 0
    #         elif not hidden and incorrect:
    #             public_score, hidden_score = 0, 0
    #             public_possible = possible
            
    #         output["tests"] += [{
    #             "name" : key + " - Public",
    #             "score" : public_score,
    #             "max_score": public_possible,
    #             "visibility": "visible",
    #             "output": repr(scores[key]["test"]) if not hidden and incorrect else "All tests passed!",
    #         }]
    #         # if not hidden and incorrect:
    #         #     output["tests"][-1]["output"] = repr(scores[key]["hint"])
            
    #         if not (not hidden and incorrect):
    #             output["tests"] += [{
    #                 "name" : key + " - Hidden",
    #                 "score" : hidden_score,
    #                 "max_score": hidden_possible,
    #                 "visibility": hidden_test_visibility,
    #                 "output": repr(scores[key]["test"]) if incorrect else "All tests passed!"
    #             }]
    #             # if hidden and incorrect:
    #             #     output["tests"][-1]["output"] = repr(scores[key]["hint"])
    
    # if config.get("show_stdout_on_release", False):
    #     output["stdout_visibility"] = "after_published"

    # if config.get("points_possible", None) is not None:
    #     output["score"] = scores["total"] / scores["possible"] * config.get("points_possible", None)

    # if config.get("score_threshold", None) is not None:
    #     if scores["total"] / scores["possible"] >= config["score_threshold"]:
    #         output["score"] = config.get("points_possible", None) or scores["possible"]
    #     else:
    #         output["score"] = 0

    with open("/autograder/results/results.json", "w+") as f:
        json.dump(output, f, indent=4)

    print("\n\n")
    # print(dedent("""\
    # Test scores are summarized in the table below. Passed tests appear as a single cell with no output.
    # Failed public tests appear as a single cell with the output of the failed test. Failed hidden tests
    # appear as two cells, one with no output (the public tests) and another with the output of the failed
    # (hidden) test that is not visible to the student.
    # """))
    df = pd.DataFrame(output["tests"])
    if "output" in df.columns:
        df.drop(columns=["output"], inplace=True)
    # df.drop(columns=["hidden"], inplace=True)
    print(df)


##################################
##### Tests for otter assign #####
##################################

import unittest
import sys
import os
import shutil
import subprocess

from subprocess import PIPE
from glob import glob
from unittest.mock import patch

from otter.argparser import get_parser
from otter.assign import main as assign
from otter.generate.token import APIClient


parser = get_parser()

TEST_FILES_PATH = "test/test-assign/"

class TestAssign(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print("\n\n\n" + ("=" * 60) + f"\nRunning {__name__}.{cls.__name__}\n" + ("=" * 60) + "\n")

    def check_gradescope_zipfile(self, path, correct_dir_path, config, tests=[], files=[]):
        # unzip the zipfile
        unzip_command = ["unzip", "-o", path, "-d", TEST_FILES_PATH + "autograder"]
        unzip = subprocess.run(unzip_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(unzip.stderr), 0, unzip.stderr.decode("utf-8"))

        self.assertDirsEqual(TEST_FILES_PATH + "autograder", correct_dir_path, ignore_ext=[])

        # cleanup
        if os.path.exists(TEST_FILES_PATH + "autograder"):
            shutil.rmtree(TEST_FILES_PATH + "autograder")

    def assertFilesEqual(self, p1, p2):
        try:
            with open(p1) as f1:
                with open(p2) as f2:
                    self.assertEqual(f1.read(), f2.read(), f"Contents of {p1} did not equal contents of {p2}")
        
        except UnicodeDecodeError:
            with open(p1, "rb") as f1:
                with open(p2, "rb") as f2:
                    self.assertEqual(f1.read(), f2.read(), f"Contents of {p1} did not equal contents of {p2}")

    def assertDirsEqual(self, dir1, dir2, ignore_ext=[]):
        self.assertTrue(os.path.exists(dir1), f"{dir1} does not exist")
        self.assertTrue(os.path.exists(dir2), f"{dir2} does not exist")
        self.assertTrue(os.path.isfile(dir1) == os.path.isfile(dir2), f"{dir1} and {dir2} have different type")

        if os.path.isfile(dir1):
            if os.path.splitext(dir1)[1] not in ignore_ext:
                self.assertFilesEqual(dir1, dir2)

        else:
            self.assertEqual(os.listdir(dir1), os.listdir(dir2), f"{dir1} and {dir2} have different contents")
            for f1, f2 in zip(os.listdir(dir1), os.listdir(dir2)):
                f1, f2 = os.path.join(dir1, f1), os.path.join(dir2, f2)
                self.assertDirsEqual(f1, f2, ignore_ext=ignore_ext)
    
    def test_convert_example(self):
        """
        Checks that otter assign filters and outputs correctly
        """
        # run otter assign
        run_assign_args = [
            "assign", "--no-run-tests", TEST_FILES_PATH + "example.ipynb", TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_assign_args)
        args.func = assign
        args.func(args)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "example-correct")

        # cleanup
        if os.path.exists(TEST_FILES_PATH + "output"):
            shutil.rmtree(TEST_FILES_PATH + "output")

    def test_otter_example(self):
        """
        Checks that otter assign filters and outputs correctly, as well as creates a correct .otter file
        """
        run_assign_args = [
            "assign", "--no-init-cell", "--no-check-all", TEST_FILES_PATH + "generate-otter.ipynb", 
            TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_assign_args)
        args.func = assign
        args.func(args)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "otter-correct")

        # cleanup
        if os.path.exists(TEST_FILES_PATH + "output"):
            shutil.rmtree(TEST_FILES_PATH + "output")      

    def test_pdf_example(self):
        """
        Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file along with PDFs
        """
        run_assign_args = [
            "assign", "--no-run-tests", TEST_FILES_PATH + "generate-pdf.ipynb", TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_assign_args)
        args.func = assign
        args.func(args)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "pdf-correct", ignore_ext=[".pdf",".zip"])
        
        # check gradescope zip file
        self.check_gradescope_zipfile(
            TEST_FILES_PATH + "output/autograder/autograder.zip", TEST_FILES_PATH + "pdf-autograder-correct",
            {
                "seed": 42,
                "show_stdout_on_release": True,
                "show_hidden_tests_on_release": True,
            },  ["q1.py","q3.py","q8.py"], ["data.csv"]
        )

        # cleanup
        if os.path.exists(TEST_FILES_PATH + "output"):
            shutil.rmtree(TEST_FILES_PATH + "output")
    
    @patch.object(APIClient,"get_token")
    def test_gradescope_example(self, mocked_client):

        """
        Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file along with PDFs.
        Additionally, includes testing Gradescope integration.
        """

        mocked_client.return_value = 'token'
        
        run_gradescope_args = [
            "assign", "--no-run-tests", "--no-check-all", TEST_FILES_PATH + "generate-gradescope.ipynb", 
            TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_gradescope_args)
        args.func = assign
        args.func(args)
        
        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "gs-correct", ignore_ext=[".pdf",".zip"])

        # check gradescope zip file
        
        self.check_gradescope_zipfile(
            TEST_FILES_PATH + "output/autograder/autograder.zip", TEST_FILES_PATH + "gs-autograder-correct",
            {
                "seed": 42,
                "show_stdout_on_release": True,
                "show_hidden_tests_on_release": True,
                "token": 'token',
                "course_id": '123',
                "assignment_id": '567',
                "filtering": True
            },  ["q1.py","q3.py","q8.py"], ["data.csv"]
        )
        # cleanup
        if os.path.exists(TEST_FILES_PATH + "output"):
            shutil.rmtree(TEST_FILES_PATH + "output")
        



###############################
##### Start Otter Service #####
###############################

MISSING_PACKAGES = False

try:
    import os
    import json
    import yaml
    import hashlib
    import jwt
    # import logging
    # import traceback
    import tornado.options
    import queries
    import stdio_proxy

    from io import BytesIO
    from datetime import datetime
    from binascii import hexlify
    from tornado.httpserver import HTTPServer
    from tornado.web import Application, RequestHandler
    from tornado.auth import GoogleOAuth2Mixin
    from tornado.ioloop import IOLoop
    from tornado.queues import Queue
    from concurrent.futures import ThreadPoolExecutor

    from ..containers import grade_assignments
    from ..utils import connect_db

    OTTER_SERVICE_DIR = "/otter-service"
    ARGS = None
    SUBMISSION_QUEUE = Queue()
    CONN = None
    EXECUTOR = ThreadPoolExecutor()

    class BaseHandler(tornado.web.RequestHandler):
        """Base login handler"""
        def get_current_user(self):
            """Gets secure user cookie for personal authentication
            """
            return self.get_secure_cookie("user")
    

    class LoginHandler(BaseHandler):
        """Default auth handler
        
        A login handler that requires instructors to setup users and passwords in database
        beforehand, allowing students to auth within the notebook.
        """
        async def get(self):
            """
            Get request for personal/default authentication login
            """
            username = self.get_argument('username', True)
            password = self.get_argument('password', True)
            pw_hash = hashlib.sha256(password.encode()).hexdigest()
            account_check = await self.db.query(
                """
                SELECT * FROM users 
                WHERE username = %s AND password = %s
                """,
                [username, pw_hash]
            )
            if len(account_check) > 0:
                print("Logged in user {} and generating API key".format(username))
                account_check.free()
                api_key = hexlify(os.urandom(32)).decode("utf-8")
                self.write(api_key)
                results = await self.db.query(
                    """
                    INSERT INTO users (api_keys, username, password) VALUES (%s, %s, %s)
                    ON CONFLICT (username)
                    DO UPDATE SET api_keys = array_append(users.api_keys, %s)
                    """,
                    [[api_key], username, pw_hash, api_key]
                )
                results.free()
            else:
                print("Failed login attempt for user {}".format(username))
                account_check.free()
                self.clear()
                self.set_status(401)
                self.finish()

        @property
        def db(self):
            return self.application.db
    

    class GoogleOAuth2LoginHandler(RequestHandler, GoogleOAuth2Mixin):
        async def get(self):
            """Google OAuth login handler

            Handler for authenticating users with Google OAuth. Requires that user sets environment
            variables containing their client key and secret. Provides users with an API key that they
            can enter in the notebook by way of authenticating.
            """
            if not self.get_argument('code', False):
                print("Redirecting user to Google OAuth")
                return self.authorize_redirect(
                    redirect_uri=self.settings['auth_redirect_uri'],
                    client_id = ARGS.google_key if ARGS.google_key else self.settings['google_oauth']['key'],
                    client_secret = ARGS.google_secret if ARGS.google_secret else self.settings['google_oauth']['secret'],
                    scope=['email', 'profile'],
                    response_type='code',
                    extra_params={'approval_prompt': 'auto'}
                )
            else:
                resp = await self.get_authenticated_user(
                    redirect_uri=self.settings['auth_redirect_uri'],
                    code=self.get_argument('code')
                )
                api_key = hexlify(os.urandom(32)).decode("utf-8")
                email = jwt.decode(resp['id_token'], verify=False)['email']
                print("Generating API key for user {} from Google OAuth".format(email))
                results = await self.db.query(
                    """
                    INSERT INTO users (api_keys, email) VALUES (%s, %s)
                    ON CONFLICT (email) 
                    DO UPDATE SET api_keys = array_append(users.api_keys, %s)
                    """,
                    [[api_key], email, api_key]
                )
                results.free()

                self.render("templates/api_key.html", key=api_key)

        @property
        def db(self):
            return self.application.db
    

    class SubmissionHandler(RequestHandler):
        """Processes and validations student submission

        Handler for processing and validating a student's submission. Ensure that notebook is present
        and valid, checks API key, and implements rate limiting to prevent spamming the autograder.
        Queues submission for grading by EXECUTOR.
        """
        async def get(self):
            """Route to warn users that this is a POST-only route"""
            self.write("This is a POST-only route; you probably shouldn't be here.")
            self.finish()

        async def post(self):
            """Post request function for handling python notebook submissions"""
            self.submission_id = None
            try:
                # check request params
                request = tornado.escape.json_decode(self.request.body)
                assert 'nb' in request.keys(), 'submission contains no notebook'
                assert 'api_key' in request.keys(), 'missing api key'

                notebook = request['nb']
                api_key = request['api_key']
                
                # run through submission
                await self.submit(notebook, api_key)
            except Exception as e:
                print(e)
            self.finish()

            # if submission successful, queue notebook for grading
            if self.submission_id is not None:
                SUBMISSION_QUEUE.put(self.submission_id)

        async def validate(self, notebook, api_key):
            """Ensures a submision is valid by checking user credentials, submission frequency, and
            validity of notebook file.

            Arguments:
                notebook (json): notebook in json form
                api_key (str): API Key generated during submission

            Returns:
                tuple: submission information
            """
            # authenticate user with api_key
            results = await self.db.query("SELECT user_id, username, email FROM users WHERE %s=ANY(api_keys) LIMIT 1", [api_key])
            user_id = results.as_dict()['user_id'] if len(results) > 0 else None
            username = results.as_dict()['username'] or results.as_dict()['email'] if len(results) > 0 else None
            results.free()
            assert user_id, 'invalid API key: {}'.format(api_key)

            # get assignment and class id from notebook metadata
            assert all(key in notebook for key in ['metadata', 'nbformat', 'cells']), 'invalid Jupyter notebook'
            assert 'assignment_id' in notebook['metadata'], 'missing required metadata attribute: assignment_id'
            assert 'class_id' in notebook['metadata'], 'missing required metadata attribute: class_id'
            assignment_id = notebook['metadata']['assignment_id']
            class_id = notebook['metadata']['class_id']

            # rate limiting
            results = await self.db.query(
                """
                SELECT timestamp 
                FROM submissions 
                WHERE user_id = %s AND assignment_id = %s AND class_id = %s
                ORDER BY timestamp DESC 
                LIMIT 1
                """, 
                (user_id, assignment_id, class_id)
            )
            last_submitted = results.as_dict()['timestamp'] if len(results) > 0 else None
            results.free()

            if last_submitted:
                delta = datetime.utcnow() - last_submitted
                # rate_limit = 120
                if delta.seconds < ARGS.rate_limit:
                    self.write_error(429, message='Please wait {} second(s) before re-submitting.'.format(ARGS.rate_limit - delta.seconds))
                    return
            
            # check that assignment exists
            results = await self.db.query("SELECT * FROM assignments WHERE assignment_id=%s LIMIT 1", [assignment_id])
            assert results, 'assignment_id {} not found on server'.format(assignment_id)
            assignment = results.as_dict()
            results.free()

            return (user_id, username, assignment['class_id'], assignment_id, assignment['assignment_name'])

        async def submit(self, notebook, api_key):
            """If valid submission, inserts notebook into submissions table in database and queues 
                it for grading.

            Arguments:
                notebook (json): notebook in json form
                api_key (str): API Key generated during submission
            """
            try:
                user_id, username, class_id, assignment_id, assignment_name = await self.validate(notebook, api_key)
            except TypeError as e:
                print("Submission failed for user with API key {}: ".format(api_key, e))
                return
            except AssertionError as e:
                print("Submission failed for user with API key {} due to due to client error: {}".format(api_key, e))
                self.write_error(400, message=e)
                return

            # fetch next submission id
            results = await self.db.query("SELECT nextval(pg_get_serial_sequence('submissions', 'submission_id')) as id")
            submission_id = results.as_dict()['id']
            results.free()

            print("Successfully received submission {} from user {}".format(submission_id, username))

            # save notebook to disk
            dir_path = os.path.join(
                self.settings['notebook_dir'],
                'class-{}'.format(class_id),
                'assignment-{}'.format(assignment_id),
                'submission-{}'.format(submission_id)
            )
            file_path = os.path.join(dir_path, '{}.ipynb'.format(assignment_name))
            if not os.path.exists(dir_path):
                os.makedirs(dir_path)
            with open(file_path, 'w') as f:
                json.dump(notebook, f)

            print("Successfully saved submission {} at {}".format(submission_id, file_path))
            
            # store submission to database
            results = await self.db.query("INSERT INTO submissions (submission_id, assignment_id, class_id, user_id, file_path, timestamp) VALUES (%s, %s, %s, %s, %s, %s)",
                                                [submission_id, assignment_id, class_id, user_id, file_path, datetime.utcnow()])
            assert results, 'submission failed'
            results.free()

            self.submission_id = submission_id

            self.write('Submission {} received.'.format(submission_id))

        @property
        def db(self):
            return self.application.db

        def write_error(self, status_code, **kwargs):
            if 'message' in kwargs:
                self.write('Submission failed: {}'.format(kwargs['message']))
            else:
                self.write('Submission failed.')


    def grade_submission(submission_id):
        """Grades a single submission with id SUBMISSION_ID

        Args:
            submission_id (str): the id of the submission to grade

        Returns:
            tuple: grading message and results dataframe for printing
        """
        global CONN
        cursor = CONN.cursor()

        cursor.execute(
            """
            SELECT user_id, submission_id, assignment_id, class_id, file_path 
            FROM submissions 
            WHERE submission_id = %s 
            ORDER BY timestamp DESC
            LIMIT 1
            """,
            (submission_id, )
        )
        user_record = cursor.fetchall()
        assert len(user_record) == 1, "Submission {} not found".format(submission_id)
        row = user_record[0]
        user_id = int(row[0])
        submission_id = int(row[1])
        assignment_id = str(row[2])
        class_id = str(row[3])
        file_path = str(row[4])

        cursor.execute(
            """
            SELECT seed
            FROM assignments
            WHERE assignment_id = %s AND class_id = %s
            """,
            (assignment_id, class_id)
        )
        assignment_record = cursor.fetchall()
        assert len(assignment_record) == 1, "Assignment {} for class {} not found".format(assignment_id, class_id)
        seed = int(assignment_record[0][0]) if assignment_record[0][0] else None

        cursor.execute(
            """
            SELECT username, email 
            FROM users 
            WHERE user_id = %s
            LIMIT 1
            """,
            (user_id, )
        )
        user_record = cursor.fetchall()
        row = user_record[0]
        username = str(row[0] or row[1])

        # Run grading function in a docker container
        stdout = BytesIO()
        stderr = BytesIO()
        try:
            with stdio_proxy.redirect_stdout(stdout), stdio_proxy.redirect_stderr(stderr):
                df = grade_assignments(
                    tests_dir=None, 
                    notebooks_dir=file_path, 
                    id=assignment_id, 
                    image=class_id + "-" + assignment_id,
                    debug=True,
                    verbose=True,
                    seed=seed
                )
                
            message = "Graded submission {} from user {}".format(submission_id, username)

            df_json_str = df.to_json()
        
            # Insert score into submissions table
            cursor.execute(
                """
                UPDATE submissions
                SET score = %s
                WHERE submission_id = %s
                """,
                (df_json_str, submission_id)
            )

        finally:
            stdout = stdout.getvalue().decode("utf-8")
            stderr = stderr.getvalue().decode("utf-8")
            with open(os.path.join(os.path.split(file_path)[0], "GRADING_STDOUT"), "w+") as f:
                f.write(stdout)
            with open(os.path.join(os.path.split(file_path)[0], "GRADING_STDERR"), "w+") as f:
                f.write(stderr)
        
        cursor.close()
        return message, df


    async def start_grading_queue(shutdown=False):
        """Pops submission ids off SUBMISSION_QUEUE and sending them into EXECUTOR to be graded
        
        Args:
            shutdown (bool): default `False`; whether or not to shutdown EXECUTOR after processing
                queue
        """
        global SUBMISSION_QUEUE

        async for submission_id in SUBMISSION_QUEUE:
            future = EXECUTOR.submit(
                grade_submission,
                submission_id
            )
            future.add_done_callback(lambda f: print(f.result()[0], "\n", f.result()[1]))

            # Set task done in queue
            SUBMISSION_QUEUE.task_done()
        
        if shutdown:
            EXECUTOR.shutdown(wait=True)


    class Application(tornado.web.Application):
        def __init__(self):
            """Initialize tornado server for receiving/grading submissions"""
            endpoint = ARGS.endpoint or os.environ.get("OTTER_ENDPOINT", None)
            assert endpoint is not None, "no endpoint address provided"
            assert os.path.isdir(OTTER_SERVICE_DIR), "{} does not exist".format(OTTER_SERVICE_DIR)
            settings = dict(
                google_oauth={
                    "key": ARGS.google_key or os.environ.get("GOOGLE_CLIENT_KEY", None), 
                    "secret": ARGS.google_secret or os.environ.get("GOOGLE_CLIENT_SECRET", None)
                },
                notebook_dir = os.path.join(OTTER_SERVICE_DIR, "submissions"),
                auth_redirect_uri = os.path.join(endpoint, "auth/callback")
            )
            handlers = [
                (r"/submit", SubmissionHandler),
                (r"/auth/google", GoogleOAuth2LoginHandler),
                (r"/auth/callback", GoogleOAuth2LoginHandler),
                (r"/auth", LoginHandler)
            ]
            tornado.web.Application.__init__(self, handlers, **settings)
            
            # Initialize database session
            self.db = queries.TornadoSession(queries.uri(
                host=ARGS.db_host,
                port=ARGS.db_port,
                dbname='otter_db',
                user=ARGS.db_user,
                password=ARGS.db_pass
            ))

except ImportError:
    # don't need requirements to use otter without otter service
    MISSING_PACKAGES = True

def main(cli_args):
    if MISSING_PACKAGES:
        raise ImportError(
            "Missing some packages required for otter service. "
            "Please install all requirements at "
            "https://raw.githubusercontent.com/ucbds-infra/otter-grader/master/requirements.txt"
        )

    global CONN
    global ARGS

    ARGS = cli_args
    CONN = connect_db(ARGS.db_host, ARGS.db_user, ARGS.db_pass, ARGS.db_port)

    port = ARGS.port
    tornado.options.parse_command_line()

    # make submissions forlder
    if not os.path.isdir(OTTER_SERVICE_DIR):
        os.makedirs(os.path.join(OTTER_SERVICE_DIR))
    
    server = HTTPServer(Application())
    server.listen(port)
    print("Listening on port {}".format(port))

    IOLoop.current().add_callback(start_grading_queue)
    IOLoop.current().start()
